#!/usr/bin/python
# change.py

'''
Classes and functions to implement landscape, demographic, and life-history
parameter change operations.
'''
#geonomics imports
from geonomics.utils import spatial as spt
from geonomics.utils import io
from geonomics.utils.viz import _check_display

#other imports
import os
import numpy as np
import numpy.random as r
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from operator import attrgetter as ag
from collections import OrderedDict as OD
from collections import Counter as C
from copy import deepcopy


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

#base class for _LandscapeChanger and _SpeciesChanger
class _Changer:
    def __init__(self, params):
        #set the type-label of the changer
        self.type = None

        #grab a deep copy of the change params
        self.change_params = deepcopy(params)

        #create self.changes, which defaults to None but will be set
        #to an iterator over all changes that are to happen to the 
        #changing object during the model (where changes in the iterator
        #are tuples of (timestep, change_function))
        self.changes = None
        #create self.next_change, which defaults to None but will be
        #set by self.set_changes and upkept by self.make_change
        self.next_change = None

    def _set_next_change(self):
        try:
            self.next_change = next(self.changes)
        except StopIteration:
            self.next_change = None

    def _make_change(self, t, additional_args, for_plotting=False,
        verbose=False):
        #if this is the right timestep for the next change
        if self.next_change is not None:
            if t == self.next_change[0] and for_plotting == False:
                if verbose:
                    print("\t**** Running the next change\t%s\n\n" % str(
                        self.next_change))
                #call the next_change function to make the change
                #NOTE: unpacking additonal_args dict into it, so that the
                #LandChanger can receive the current Landscape object, rather
                #than capturing the Landscape object within the closure, taking
                #up memory, or capturing a pointer to the object and risking
                #problems that I'm not foreseeing (a la https://docs.python.
                #org/3/faq/programming.html#why-do-lambdas-defined-in-a
                #-loop-with-different-values-all-return-the-same-result),
                #and so that SpeciesChanger can similarly get the current
                #Species object
                self.next_change[1](changer = self, **additional_args)
                    #NOTE: feeding self in so that some change functions
                    #that need to set Changer attributes (e.g. spp-changer
                    #demographic functions set the base_K raster at
                    #the beginning of a period of dem changes)
                    #have access to the Changer object
                #then load the next change after that
                self._set_next_change()
                #then recurse this function, so that multiple changes
                #will happen this timestep if need be
                self._make_change(t, additional_args)

    #add a change (in the form of a (timestep, change_fn) pair) at the
    #appropriate timestep
    def _add_change(self, change):
        changes = [self.next_change]
        more = True
        while more:
            try:
                changes.append(next(self.changes))
            except StopIteration:
                more = False
        before = [c[0] <= change[0] for c in self.changes]
        insert = before.index(False)
        changes = changes[:insert] + [change] + changes[insert:]
        self.next_change = changes[0]
        self.changes = iter(changes)


class _LandscapeChanger(_Changer):
    def __init__(self, land, land_change_params, mod):
        super(_LandscapeChanger, self).__init__(land_change_params)
        #set the type-label of the changer
        self.type = 'land'

        #create an empty dict to hold all the basic change info
        #(end_rast, timesteps) for each layer that will change (so that
        #other objects, such as movement_surfaces, can access this
        #information and use it to build their own change-series)
        self.change_info = {}

    #call self.set_changes() to set self.changes and self.next_change
        self._set_changes(land)

    #method to set the changes stipulated in land_change_params dict for
    #the land object
    def _set_changes(self, land):

        #get the time-series of layers for each lyr_num
        lyrs = {}
        surfs = {}
        for lyr_num in self.change_params.keys():
            #get the conglomerate lyr series, across all change events
            #parameterized for this Layer
            conglom_lyr_series = _make_conglom_lyr_series(land, lyr_num,
                self.change_params[lyr_num])
            #add the conglomerate lyr series to the lyrs object
            lyrs[lyr_num] = conglom_lyr_series
            #set the _LandscapeChanger.change_info attribute, so that it can be
            #grabbed later for building _move_surf,
            #_disp_surf, and other objects
            self.change_info[lyr_num] = {**self.change_params[lyr_num]}

        #get everything in chronological order (with lyr-change functions
        #coming before surf-change functions if they're in the same timestep)
        lyr_changes = []
        for lyr_num in lyrs.keys():
            lyr_changes.extend([(t, lyr_num,
                                   lyr) for t,lyr in lyrs[lyr_num]])
        lyr_changes = sorted(lyr_changes, key = lambda x: x[0])

        #make a list of change_fns for the changes in all_changes, and make
        #self.changes an iterator over that list
        change_fns = [(change[0],
            _get_lyr_change_fn(*change[1:])) for change in lyr_changes]
        change_fns = iter(change_fns)
        self.changes = change_fns
        #load the first change to happen into the self.next_change attribute
        self._set_next_change()


class _SpeciesChanger(_Changer):
    def __init__(self, spp, spp_change_params, land):
        super(_SpeciesChanger, self).__init__(spp_change_params)
        self.type = 'spp'

        #an attribute that is used by some dem-change fns, as a baseline
        #population size at the start of the demographic change event
        #against which later timesteps' changes are defined
        self.base_K = None

    #call self._set_changes() to set self.changes and self.next_change
        self._set_changes(spp, land)

    #method to set the base_K attribute to spp.K
    def _set_base_K(self, spp):
        self.base_K = spp.K

    #method to set the changes stipulated in params dict for the spp object
    def _set_changes(self, spp, land):
        #pull out the parts of the params
        try:
            dem_change_params = self.change_params.dem
        except Exception:
            dem_change_params = None
        try:
            parameter_change_params = self.change_params.life_hist
        except Exception:
            parameter_change_params = None
        #check if this spp has a _move_surf, and if there are land-changes
        #that affect its lyr
        move_surf_change_fns = []
        if (spp._move_surf is not None
            and land._changer is not None
            and spp._move_surf.lyr_num in land._changer.change_info.keys()):
            #if so, grab that lyr's land change params and create a
            #move_surf_series with them, to be 
            #added to this spp's change fns
            lc_move_surf_params = land._changer.change_info[
                                                    spp._move_surf.lyr_num]
            #create a time-series of movement surfaces 
            surf = spp._move_surf
            surf_series = _make_conductance_surface_series(land,
                start_lyr = land[surf.lyr_num], mixture = surf.mix,
                kappa = surf.kappa, approx_len = surf.approx_len,
                change_params_one_lyr = lc_move_surf_params)
            #and create change fns from them
            move_surf_change_fns.extend(
                _get_conductance_surface_change_fns(surf_series))

        #check if this spp has a _disp_surf, and if there are land-changes
        #that affect its lyr
        disp_surf_change_fns = []
        if (spp._disp_surf is not None
            and land._changer is not None
            and spp._disp_surf.lyr_num in land._changer.change_info.keys()):
            #if so, grab that lyr's land change params and create a
            #disp_surf_series with them, to be 
            #added to this spp's change fns
            lc_disp_surf_params = land._changer.change_info[
                                                    spp._disp_surf.lyr_num]
            #create a time-series of movement surfaces 
            surf = spp._disp_surf
            surf_series = _make_conductance_surface_series(land,
                start_lyr = land[surf.lyr_num], mixture = surf.mix,
                kappa = surf.kappa, approx_len = surf.approx_len,
                change_params_one_lyr = lc_disp_surf_params)
            #and create change fns from them
            disp_surf_change_fns.extend(
                _get_conductance_surface_change_fns(surf_series))


        #set the demographic changes, if applicable
        dem_change_fns = []
        if dem_change_params is not None:
            for event, event_params in dem_change_params.items():
                if True in [v is not None for v in event_params.values()]:
                    dem_change_fns.extend(_get_dem_change_fns(spp,
                                                             **event_params))

        #set the other changes, if applicable
        parameter_change_fns = []
        if parameter_change_params is not None:
            for parameter, parameter_params in parameter_change_params.items():
                if True in [v is not None for v in parameter_params.values()]:
                    parameter_change_fns.extend(
                        _get_parameter_change_fns(parameter,
                        **parameter_params))

        #put all the changes in chronological order
        if (len(dem_change_fns) + len( parameter_change_fns) + len(
            move_surf_change_fns) + len(disp_surf_change_fns) > 0):
            change_fns = dem_change_fns + parameter_change_fns
            change_fns = change_fns + move_surf_change_fns
            change_fns = change_fns + disp_surf_change_fns
            change_fns = sorted(change_fns, key = lambda x: x[0])

            #make self.changes an iterator over that list
            change_fns = iter(change_fns)
            self.changes = change_fns
            #load the first change to happen into the self.next_change
            #attribute
            self._set_next_change()

    #a method to visualize the species changes that will occur
    def _plot_dem_changes(self, spp):
        if self.next_change is None:
            print(("No demographic changes remaining for this species. "
                "They were likely already run."))
        else:
            cop_spp = deepcopy(spp)
            cop_self = deepcopy(self)
            cop_changes = deepcopy(cop_self.changes)
            step_list = [cop_self.next_change[0]]
            more = True
            while more:
                try:
                    next_change = next(cop_changes)
                    step_list.append(next_change[0])
                except StopIteration:
                    more = False
            end = int(1.1*max(step_list))
            #set spp.K to 1, for viz purposes
            spp.K = 1
            #and set cop_self.base_K 
            cop_self._set_base_K(spp)
            Ks = []
            for t in range(end):
                cop_self._make_change(t, for_plotting = True,
                                      additional_args={'spp': spp})
                Ks.append(spp.K)
            plt.plot(range(end), Ks)
            #set spp back to its original value
            spp = cop_spp


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

    #########################
    # for _LandscapeChanger #
    #########################

# function that takes a starting rast, its layer, an ending rast, a few
# timestep arguments, and returns a linearly interpolated stack of rasters
def _make_lyr_series(lyr, change_rast, start_t, end_t, n_steps,
                     coord_prec=0):
    start_rast = lyr.rast
    dim = lyr.dim
    scale_min = lyr._scale_min
    scale_max = lyr._scale_max

    #get (rounded) evenly spaced timesteps at which to implement the changes
    timesteps = np.int64(np.round(np.linspace(start_t, end_t, n_steps)))

    #if change_rast is a file or an array
    if ((isinstance(change_rast, str) and os.path.isfile(change_rast))
        or isinstance(change_rast, np.ndarray)):
        #if it's an array, keep it, and set the prj, res, dim, ulc values;
        #if it's a file, read it in and get the dim, res, ulc, and prj values
        if isinstance(change_rast, np.ndarray):
            #NOTE: this resets dim from the argument that was fed in
            dim = change_rast.shape
            # NOTE: setting ulc, res, and prj to the same values as the layer
            # provided to the layer argument, so that the None values don't
            # throw an error in line 515's assertion below, but in reality
            # there might be a better, 'safer' way to handle this
            ulc = lyr.ulc
            res = lyr.res
            prj = lyr.prj
        elif os.path.isfile(change_rast):
            #NOTE: this resets dim from the argument that was fed in
            change_rast, dim, res, ulc, prj = io._read_raster(change_rast,
                                                              coord_prec,
                                                              dim=dim)
            # scale the raster correctly
            change_rast, scale_min_out, scale_max_out = spt._scale_raster(
                                            change_rast, scale_min, scale_max)
            assert (scale_min == scale_min_out
                    and scale_max == scale_max_out), ("The scale_min and "
                                                      "scale_max values "
                                                      "returned from scaling "
                                                      "the input change "
                                                      "raster don't match "
                                                      "those of Layer, to "
                                                      "which the raster "
                                                      "corresponds.") % lyr.idx
        #flatten the start and end rasters
        start = start_rast.flatten()
        end = change_rast.flatten()
        #get a column of values for each grid cell on the landscape, linearly
        #spaced between that cell's start and end values
        #NOTE: linspace(..., ..., n+1)[1:] gives us the changed rasters
        #for n steps, leaving off the starting lyr value because that's
        #already the existing lyr so we don't want that added into our changes
        rast_series = np.vstack([np.linspace(start[i],
            end[i], n_steps+1)[1:] for i in range(len(start))])
        #then reshape each timeslice into the dimensions of start_rast
        rast_series = [rast_series[:,i].reshape(
            start_rast.shape) for i in range(rast_series.shape[1])]

    #elif it's a directory
    elif isinstance(change_rast, str) and os.path.isdir(change_rast):
        files = os.listdir(change_rast)
        assert len(files) == n_steps, ("The number of files in the directory "
            "provided for the 'change_rast' parameter is not equal to the "
            "number provided for the 'n_steps' parameter.")
        #check that each file starts with an integer
        for f in files:
            assert os.path.splitext(f.split('_')[0])[0].isnumeric(), ("In the "
                "directory provided for the "
                "'change_rast' parameter, the file %s does not start "
                "with an integer followed by an underscore. This is "
                "necessary, to indicate the timestep during the Landscape "
                "change event at which the file's raster should "
                "be used.") % f
        #ensure that the files are ordered by the beginning integer
        #NOTE: should already be sorted by natural sorting, but just in case
        steps_and_files = {int(
            os.path.splitext(f.split('_')[0])[0]):f for f in files}
        files = [steps_and_files[i] for i in sorted(steps_and_files.keys())]
        #read in the whole series, check all prj, dim, ulc, and res values
        #are equal, then set the prj, dim, ulc, and res values, and
        #set rasters equal to rast_series
        all_rasts = [io._read_raster(os.path.join(
            change_rast, f), coord_prec, dim) for f in files]
        dim_cts = C([i[1] for i in all_rasts])
        assert len(dim_cts) == 1, ("The dimensions of "
            "all files in the directory provided for the 'change_rast' "
            "parameter are not equal. Most files have dimensions %s, but the "
            "following files differ: %s.") % (
            str([k for k,v in dim_cts.items() if v == max(dim_cts.values())]),
            str([files[k] for k,v in dim_cts.items() if v != max(
            dim_cts.values())]))
        #NOTE: this resets dim from the argument that was fed in
        dim = all_rasts[0][1]
        res_cts = C([tuple([*i[2]]) for i in all_rasts])
        assert len(res_cts) == 1, ("The spatial resolutions of "
            "all files in the directory provided for the 'change_rast' "
            "parameter are not equal. Most files have resolution %s, but the "
            "following files differ: %s.") % (
            str([k for k,v in res_cts.items() if v == max(res_cts.values())]),
            str([files[k] for k,v in res_cts.items() if v != max(
            res_cts.values())]))
        res = all_rasts[0][2]
        ulc_cts = C([tuple([*i[3]]) for i in all_rasts])
        assert len(ulc_cts) == 1, ("The upper left corners of "
            "all files in the directory provided for the 'change_rast' "
            "parameter are not equal. Most files have upper left corner %s, "
            "but the following files differ: %s.") % (
            str([k for k,v in ulc_cts.items() if v == max(ulc_cts.values())]),
            str([files[k] for k,v in ulc_cts.items() if v != max(
            ulc_cts.values())]))
        ulc = all_rasts[0][3]
        prj_cts = C([i[4] for i in all_rasts])
        assert len(prj_cts) == 1, ("The projections of "
            "all files in the directory provided for the 'change_rast' "
            "parameter are not equal. Most files have projection %s, but the "
            "following files differ: %s.") % (
            str([k for k,v in prj_cts.items() if v == max(prj_cts.values())]),
            str([files[k] for k,v in prj_cts.items() if v != max(
            prj_cts.values())]))
        prj = all_rasts[0][4]
        rast_series = [i[0] for i in all_rasts]
        # rescale the rasters
        rasts_and_scale_vals = [spt._scale_raster(
                        rast, scale_min, scale_max) for rast in rast_series]
        assert len(set([i[1] for i in rasts_and_scale_vals])) == 1, (
            "More than one unique scale_min value was returned from rescaling "
            "the directory of rasters provided for Layer %i.") % lyr.idx
        assert len(set([i[2] for i in rasts_and_scale_vals])) == 1, (
            "More than one unique scale_max value was returned from rescaling "
            "the directory of rasters provided for Layer %i.") % lyr.idx
        scale_min_out = rasts_and_scale_vals[0][1]
        scale_max_out = rasts_and_scale_vals[0][2]
        assert (scale_min == scale_min_out
                and scale_max == scale_max_out), ("The scale_min and "
                                                  "scale_max values "
                                                  "returned from scaling "
                                                  "the input change "
                                                  "rasters don't match "
                                                  "those of Layer %i, to "
                                                  "which the raster "
                                                  "corresponds.") % lyr.idx
        rast_series = [i[0] for i in rasts_and_scale_vals]
        #try to get the timesteps from the filenames
        try:
            timesteps = [int(os.path.splitext(
                i.split('_')[0])[0]) for i in files]
        except Exception as e:
            raise ValueError("Unable to extract timesteps from the beginning "
                "of all filenames in the directory provided to the "
                "'change_rast' parameter. Each filename must begin with an "
                "integer indicating the timestep at which that raster should "
                "be switched out for the Layer's previous value, followed "
                "by an underscore (e.g. '50_mat_2011.tif' for a raster that "
                "should be used at timestep 50).")
        #check that there are no repeat timesteps
        timestep_cts = C(timesteps)
        assert len(set(timestep_cts.values())) == 1, ("Not all timesteps "
            "indicated by the beginnings of the files in the directory "
            "provided for the 'change_rast' parameter are unique. "
            "Duplicated timesteps include: %s") % (
            str([k for k, v in timestep_cts if v > 1]))
        #check that first and last timesteps of the files equal start_t & end_t
        assert timesteps[0] == start_t, ("The timesteps indicated by the "
            "beginnings of the filenames must start at the timestep indicated "
            "by the 'start_t' parameter. The 'start_t' parameter was provided "
            "the value %i, but the first timestep found in the directory was "
            "%i (from the file named '%s').") % (start_t, timesteps[0],
            files[0])
        assert timesteps[::-1][0] == end_t, ("The timesteps indicated by the "
            "beginnings of the filenames must end at the timestep indicated "
            "by the 'end_t' parameter. The 'end_t' parameter was provided "
            "the value %i, but the last timestep found in the directory was "
            "%i (from the file named '%s').") % (end_t, timesteps[::-1][0],
            files[::-1][0])

    #otherwise raise a ValueError
    else:
        raise ValueError(("The value provided for the 'change_rast' parameter "
            "must be either a numpy.ndarray or a path to a valid file "
            "(indicating the endpoint-raster of the change event), or a "
            "path to a directory of valid files (one for each timestep in "
            "the change event)."))

    #check that all the lengths match up
    assert len(rast_series) == n_steps, ("The length of the rast_series "
        "variable is not equal to the n_steps variable.")
    assert len(rast_series) == len(timesteps), ("The number of changing "
        "rasters created is not the same as the number of timesteps to be "
        "assigned to them")

    #zip the timesteps and the rasters together and return them as a list
    rast_series = list(zip(timesteps, rast_series))

    return(rast_series, dim, res, ulc, prj)


#function for getting the congolmerate lyr_series for all change
#events parameterized for a Layer
def _make_conglom_lyr_series(land, lyr_num, change_params_one_lyr):
    #check that the timeframes of the change events for this layer
    #don't overlap 
    all_timesteps = [[*range(v.start_t,
        v.end_t)] for v in change_params_one_lyr.values()]
    all_timesteps = [t for i in all_timesteps for t in i]
    assert (len(set(all_timesteps)) == len(all_timesteps)), ("Some "
        "of the change events for Layer number %i overlap "
        "in time.") % lyr_num
    #get the congolmerated lyr_series for all change events for this
    #layer
    all_lyr_series = [_make_lyr_series(land[lyr_num],
                                       coord_prec=land[lyr_num].coord_prec,
                                **v) for v in change_params_one_lyr.values()]
    #check that dim, res, ulc, and prj values are the same for each
    #change event
    assert len(set([i[1] for i in all_lyr_series]))==1, (""
        "Not all change events for Layer number %i produce layer "
        "series with the same dimensions.  Resulting dimensions "
        "include: (%s).") % (lyr_num, ', '.join([*set(
        [i[1] for i in all_lyr_series])]))
    assert len(set([tuple([*i[2]]) for i in all_lyr_series]))==1, (""
        "Not all change events for Layer number %i produce layer "
        "series with the same resolution.  Resulting resolutions "
        "include: (%s).") % (lyr_num, ', '.join([*set(
        [i[2] for i in all_lyr_series])]))
    assert len(set([tuple([*i[3]]) for i in all_lyr_series]))==1, (""
        "Not all change events for Layer number %i produce layer "
        "series with the same upper-left corners.  Resulting "
        "upper-left corners include: (%s).") % (lyr_num, ', '.join([*set(
        [i[3] for i in all_lyr_series])]))
    assert len(set([i[4] for i in all_lyr_series]))==1, (""
        "Not all change events for Layer number %i produce layer "
        "series with the same projection.  Resulting projections "
        "include: (%s).") % (lyr_num, ', '.join([*set(
        [i[4] for i in all_lyr_series])]))
    #take dim, res, ulc, and prj as single values
    dim_res_ulc_prj_list = [[i[n] for i in all_lyr_series] for n in range(1,5)]
    # NOTE: next line just coerces ndarrays to tuples, so that they can be
    # hashed into a set in the following line
    dim_res_ulc_prj_list = [[tuple([*i]) for i in item] if isinstance(
        item[0], np.ndarray) else item for item in dim_res_ulc_prj_list]
    dim, res, ulc, prj = [[*set(item)][0] for item in dim_res_ulc_prj_list]
    #check that dim, res, ulc, and prj match that of the land
    assert dim == land.dim, ('Dimensionality of lyr_series fed '
        'into _LandscapeChanger for Layer number %i does not '
        'match that of the land to be changed.') % lyr_num
    assert res == land.res or res is None, ('Resolution of '
        'lyr_series fed into _LandscapeChanger for Layer number %i '
        'does not match that of the land to be changed.') % lyr_num
    assert ulc == land.ulc or ulc is None, ('Upper-left corner of '
        'lyr_series fed into _LandscapeChanger for Layer number %i '
        'does not match that of the land to be changed.') % lyr_num
    assert prj == land.prj or prj is None, ('Projection of '
        'lyr_series fed into _LandscapeChanger for Layer number %i '
        'does not match that of the land to be changed.') % lyr_num
    #combine all lyr_series into one single lyr series
    conglom_lyr_series = [lyr_step for series in [
        i[0] for i in all_lyr_series] for lyr_step in series]
    return conglom_lyr_series


#function that takes a Landscape, a lyr_num and a dictionary of
#{t_change:new_lyr} and creates an _LandscapeChanger object that will change
#out Landscape[lyr_num] for new_lyr at each requisite t_change timestep
def _make_custom_lyr_series(lyr_dict):
    pass


def _get_lyr_change_fn(lyr_num, new_lyr_rast):
    def fn(changer, land, lyr_num = lyr_num, new_lyr_rast = new_lyr_rast):
        land._set_raster(lyr_num, new_lyr_rast)
    return(fn)


    #######################
    # for _SpeciesChanger #
    #######################
def _make_conductance_surface_series(land, start_lyr, mixture, kappa,
        approx_len, change_params_one_lyr):
    #get the time-series of lyrs across the land-change event (can
    #be at a reduced temporal resolution determined by t_res_reduct_factor,
    #because probably unnecessary to change the move_surf or disp_surf every
    #time the land changes, and they could be a fairly large objects to
    #hold in memory anyhow; but t_res_reduct_factor defaults to 1)
    conglom_lyr_series= _make_conglom_lyr_series(land, start_lyr.idx,
        change_params_one_lyr)
    #then get the series of _ConductanceSurface objects
    surf_series = []
    dummy_lyr = deepcopy(start_lyr)
    for t, rast in conglom_lyr_series:
        #change the lyr's raster
        dummy_lyr.rast = rast
        #then create a Movement_Surface using the copied Landscape
        surf_series.append((t, spt._ConductanceSurface(dummy_lyr, mixture,
            approx_len = approx_len, vm_distr_kappa = kappa)))
    return(surf_series)


def _get_conductance_surface_change_fns(surf_series):
    timesteps = []
    fns = []
    for t, new_surf in surf_series:
        def fn(changer, spp, new_surf = new_surf):
            spp._move_surf = new_surf
        timesteps.append(t)
        fns.append(fn)
    change_fns = zip(timesteps, fns)
    return(change_fns)

        ###############
        # demographic #
        ###############

def _get_dem_change_fns(spp, kind, start_t=None, end_t=None, rate=None,
    interval=None, n_cycles=None, size_range=None,
    distr='uniform', min_size=None, max_size=None, timesteps=None,
                                    sizes=None, increase_first=True):
    if kind == 'monotonic':
        fns = _get_monotonic_dem_change_fns(rate = rate,
            start_t = start_t, end_t = end_t)
    elif kind == 'stochastic':
        fns = _get_stochastic_dem_change_fns(start_K = spp.K,
            start_t = start_t, end_t = end_t, interval = interval,
            size_range = size_range, distr = distr)
    elif kind == 'cyclical':
        fns = _get_cyclical_dem_change_fns(start_t = start_t,
            end_t = end_t, n_cycles = n_cycles, size_range = size_range,
            min_size = min_size, max_size = max_size,
                                        increase_first = increase_first)
    elif kind == 'custom':
        fns = _get_custom_dem_change_fns(timesteps = timesteps, sizes = sizes)
    return(fns)


def _make_dem_change_fns(sizes, timesteps, K_mode='base'):
    fns = []
    if K_mode == 'current':
        for size in sizes:
            def fn(changer, spp, size = size):
                spp.K*=size
            fns.append(fn)
    elif K_mode == 'base':
        t0 = timesteps[0]
        for size in sizes:
            def fn(changer, spp, size=size, t0 = t0):
                if spp.t == t0:
                    changer._set_base_K(spp)
                spp.K = changer.base_K*size
            fns.append(fn)
    change_fns = list(zip(timesteps, fns))
    return(change_fns)


#will generate exponential change in the population size by iteratively
#multiplying a carrying-capacity (K) raster
def _get_monotonic_dem_change_fns(rate, start_t, end_t):
    #get the timesteps for the demogprahic changes
    #NOTE: setting start_t and end_t to the same value will create a
    #single-timestep change (e.g. a sudden bottleneck or rapid expansion)
    timesteps = range(start_t, end_t+1)
    sizes = [rate]*len(timesteps)
    change_fns = _make_dem_change_fns(sizes, timesteps, K_mode='current')
    return(change_fns)


#NOTE: should I provide an option for linear monotonic change (because I can
#set pc.base_K and then multiply it by rate*t at each t of a period of change)?


#will create stochastic changes around a baseline carrying-capacity (K) raster
def _get_stochastic_dem_change_fns(start_K, size_range, start_t, end_t,
    interval, distr = 'uniform'):
    if interval is None:
        interval = 1
    timesteps = range(start_t, end_t + 1, interval)
    if distr == 'uniform':
        sizes = r.uniform(*size_range, len(timesteps))
    elif distr == 'normal':
        mean = np.mean(size_range)
        sd = (size_range[1] - size_range[0])/6
        sizes = r.normal(loc = mean, scale = sd, size = len(timesteps))
    else:
        raise ValueError(("Argument 'distr' must be a value among "
                                            "['uniform', 'normal']"))
    #make size return to starting size
    sizes[-1] = 1
    #make all the change functions
    change_fns = _make_dem_change_fns(sizes, timesteps, K_mode = 'base')
    return(change_fns)


def _get_cyclical_dem_change_fns(start_t, end_t, n_cycles,
    size_range=None, min_size=None, max_size=None, increase_first=True):
    #detemine the min and max sizes for the cycles, based on input arguments
    if size_range is not None and min_size is None and max_size is None:
        min_size, max_size = size_range
    elif size_range is None and min_size is not None and max_size is not None:
        pass
    #and throw an informative error if both size_range and min_size&max_size
    #arguments are provided (or neither)
    else:
        raise ValueError(('Must either provide size_range (as a tuple of '
            'minimum and maximum sizes), or provide min_size and max_size '
                                                'separately, but not both.'))
    #check that there are enough timesteps to complete the cycles
    assert n_cycles <= (end_t - start_t)/2, ('The number of cycles requested must '
        'be no more than half the number of time steps over which the cycling '
                                                        'should take place.')
    #create a base cycle (one sine cycle)
    base = np.sin(np.linspace(0, 2*np.pi,1000))
    #flip it if cycles should decrease before increasing
    if not increase_first:
        base = base[::-1]
    #scale the base cycle to the min_size
    scaled_base = [1 + n*(max_size-1) if n>=0 else n for n in base]
    #and scale it to the max_size
    scaled_base = [1 + n*(1-min_size) if n<0 else n for n in scaled_base]
    scaled_base = np.array(scaled_base)
    #then get the timesteps at which each cycle should complete
    #NOTE: subtract 1 to imitate Python range stop-exclusivity
    cycle_timesteps = np.int32(np.linspace(start_t, end_t, n_cycles+1))
    #get from that the length in timesteps of each cycle
    cycle_lengths = np.diff(cycle_timesteps)
    #get from that the pop sizes at each timestep (which will approximate
    #the scaled sine curve from scaled_base
    sizes = np.hstack([scaled_base[np.int32(np.linspace(1, len(scaled_base)-1,
                                            l))] for l in cycle_lengths] + [1])
    #then get all timesteps across the cycles, to match up to the pop sizes
    timesteps = range(cycle_timesteps[0], cycle_timesteps[-1]+1)
    #get all the change functions for those timesteps
    change_fns = _make_dem_change_fns(sizes, timesteps, K_mode = 'base')
    return(change_fns)


def _get_custom_dem_change_fns(timesteps, sizes):
    assert len(timesteps) == len(sizes), ('For custom demographic changes, '
                    'timesteps and sizes must be iterables of equal length.')
    change_fns = _make_dem_change_fns(sizes, timesteps, K_mode = 'base')
    return(change_fns)


        #################
        #   life_hist   #
        #################

def _make_parameter_change_fns(parameter, timesteps, vals):
    fns = []
    for val in vals:
        def fn(changer, spp, parameter = parameter, val = val):
            setattr(spp, parameter, val)
        fns.append(fn)
    change_fns = list(zip(timesteps, fns))
    return(change_fns)


def _get_parameter_change_fns(parameter, timesteps, vals):
    assert len(timesteps) == len(vals), ("For custom changes of the '%s' "
        "parameter, timesteps and vals must be iterables of equal "
        "length.") % parameter
    change_fns = _make_parameter_change_fns(parameter, timesteps, vals)
    return(change_fns)




#!/usr/bin/python
# genome.py

# flake8: noqa


'''
Classes, associated methods, and supporting functions for all genomic
components
'''

# geonomics imports
from geonomics.ops import mutation
from geonomics.structs.individual import Individual
from geonomics.utils.viz import _check_display

# other imports
import numpy as np
import pandas as pd
from numpy import random as r
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from collections import OrderedDict as OD
import warnings
import random
import bisect
import bitarray
import msprime
import tskit


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################


# an error to raise if a simulation is not given enough neutral loci to fit
# the total number of expected mutations over the total parameterized runtime
# under the infinite sites model
class MutationRateError(Exception):
    pass


class Recombinations:
    """
    Container for the parameters and data necessary to simulate recombination.

    Not intended for public use.
    """
    def __init__(self, L, positions, n, r_distr_alpha, r_distr_beta,
                 recomb_rates, jitter_breakpoints):
        # define the bitarray unit corresponding to each homologue
        self._bitarray_dict= {0: '10', 1: '01'}

        # genome length
        self._L = L
        # organize the potential recombination breakpoint positions
        if positions is None:
            positions = np.arange(self._L)
        else:
            positions.sort()
        positions = [*positions]
        # optimize the data-type, to save some memory
        if len(positions) <= 2**16:
            self._positions = np.int16(np.array(positions))
        else:
            self._positions = np.int32(np.array(positions))

        # number of recombination events to simulate and cache for the model
        # NOTE: the higher this number, the more accurately geonomics
        #       will simulate the stipulated recombination rates
        self._n = n
        # alpha and beta parameters for the beta distribution from which
        # recombination rates can be drawn
        self._r_distr_alpha = r_distr_alpha
        self._r_distr_beta = r_distr_beta
        # flag indicating whether or not breakpoint locations should be
        # jittered slightly off of their x.5 defaults, so that
        # tskit.TreeSequence will correctly track and report the number of trees
        self._jitter_breakpoints = jitter_breakpoints
        # get the recombination rates
        if recomb_rates is not None:
            assert len(recomb_rates) == len(positions), ("Lengths of provided "
                                                         "recombination "
                                                         "rates and "
                                                         "recombination "
                                                         "positions don't "
                                                         "match!")
            assert recomb_rates[0] == 0, ("The first recombination rate (i.e. "
                                          "index 0) must be 0 (because no "
                                          "recombination can be modeled as "
                                          "taking place prior to the first "
                                          "locus being simulated).")
            self._rates = recomb_rates
        else:
            self._rates = self._draw_recombination_rates()

    def _set_events(self, use_subsetters, nonneutral_loci, use_tskit):
        # get the potential recombination breakpoints, their recombination
        # rates, and the cache of simulated recombination events to be used
        # by the model
        # NOTE: each item will only be generated and stored if it is needed,
        #       depending on whether or not tskit is being used and, if it is
        #       being used, whether or not there are non-neutral loci whose
        #       genotypes need to be tracked
        breakpoints, subsetters = self._draw_recombination_events(
                                            use_subsetters=use_subsetters,
                                            nonneutral_loci=nonneutral_loci,
                                            use_tskit=use_tskit)
        # if using tskit, set seg_info and use the breakpoints dict
        if use_tskit:
            self._breakpoints = breakpoints
            self._set_seg_info()
        # otherwise don't set seg_info and make the breakpoints dict None
        else:
            self._breakpoints = None
        # either way, set the subsetters
        self._subsetters = subsetters

    def _get_events(self, size):
        events = random.sample(self._events, size)
        return events

    def _get_subsetter(self, event_key):
        return self._subsetters[event_key]


    # update all the recombination subsetters in accord with a new non-neutral
    # mutation
    def _update_subsetters(self, mutation_loc, genotype_arr_mutation_idx):
        for k in self._subsetters.keys():
            # determine whether the number of crossovers that has happened up
            # to the mutation's locus has been even (in which case the
            # recomb path is 'back' on homologue 0) or odd (in which
            # case it is on homologue 1)
            # NOTE: this is because all recomb paths start on homologue 0,
            #       then are used either for subsetting starting from either
            #       homol 0 or homol 1 on the fly in mating.py using the 
            #       variable start_homologues
            #       in function _do_mating_singl_offspring
            subsetter_homologue = bisect.bisect_left(self._breakpoints[k],
                                                     mutation_loc) % 2
            # use the bitarray_dict to get the 2-digit binary string
            # corresponding to the homologue that the recomb path is on
            # when it reaches mutation_loc, to be inserted into the subsetter
            # in the appropriate location
            subsetter_insert = self._bitarray_dict[subsetter_homologue]
            # find the insertion-point index of the current subsetter 
            # (just twice the non-neutral genotype array's index)
            subsetter_idx = genotype_arr_mutation_idx * 2
            # create the new subsetter by concatenating:
            #  1. the original subsetter, up to but excluding the insertion pt
            #  2. the insert
            #  3. the remainder of the original subsetter
            new_subsetter = (self._subsetters[k][:subsetter_idx] +
                subsetter_insert + self._subsetters[k][subsetter_idx:])
            self._subsetters[k] = new_subsetter


    # draw recombination rates as needed according to parameter values
    def _draw_recombination_rates(self):
        if (self._r_distr_alpha is not None and self._r_distr_beta is not None):
            recomb_rates = np.clip(np.random.beta(a=self._r_distr_alpha,
                                                  b=self._r_distr_beta,
                                                  size=len(self._positions)),
                                   a_min=0, a_max=0.5)
        elif self._r_distr_alpha is not None:
            recomb_rates = np.ones(len(self._positions)) * self._r_distr_alpha
        else:
            #NOTE: if no alpha and beta values were provided for the recomb-rate
            # beta distribution, set all interlocus recomb rates to a value that
            # results in an average of one recombination per gamete produced
            homog_recomb_rate = 1 / self._L
            recomb_rates = np.ones(len(self._positions)) * homog_recomb_rate
        # set the 0th rate to 0 (because this will ensure that all recombination
        # paths start on homologue 0, such that when the start homologue of a
        # certain gamete equals 1 and its genome array is L-R flipped before
        # subsetting, the effective start homologue of the new, subsetted gamete
        # will actually be 1)
        recomb_rates[0] = 0
        return recomb_rates


    # simulate recombination events
    def _draw_recombination_events(self, use_subsetters, nonneutral_loci,
                                   use_tskit):
        """
        NOTE: Positions and recomb_rates must be provided already sorted!
        """
        # create the breakpoints dict
        recombinations = [r.binomial(1, self._rates) for _ in range(self._n)]
        breakpoints = [self._positions[np.where(
                                        recomb)] for recomb in recombinations]
        for bp in breakpoints:
            if len(bp) > 0:
                assert 0 not in bp, '%s' % str(bp)
        # recast it as an int-keyed dict, so that when I use the recombination
        # events to simulate recombination I don't have to pass around the long
        # arrays, but instead just random keys to index them out on the fly
        breakpoints = dict(zip(range(len(breakpoints)), breakpoints))

        if use_subsetters:
            # get the recombination paths used to make the bitarray subsetters
            if ((use_tskit and len(nonneutral_loci) > 0)
                or not use_tskit):
                recomb_paths = [np.cumsum(
                                    recomb) % 2 for recomb in recombinations]
                # if using tskit, grab just the values at the nonneutral loci
                # (because the subsetters will be used only for subsetting
                # the non-neutral genotypes carried with the individuals)
                if use_tskit:
                    paths = [path[nonneutral_loci] for path in recomb_paths]
                # otherwise, use all values
                else:
                    paths = [path for path in recomb_paths]

                subsetters = [bitarray.bitarray(''.join([self._bitarray_dict[
                            step] for step in path])) for path in paths]
                # recast as integer-indexed dict
            else:
                subsetters = [bitarray.bitarray() for _ in range(
                                                        len(recombinations))]
            subsetters = dict(zip(range(len(subsetters)), subsetters))
            assert len(breakpoints) == len(subsetters)
        else:
            subsetters = None
        return (breakpoints, subsetters)


    # calculate the left and right segment edges for each recombination event
    def _set_seg_info(self):
        seg_info = {}
        ##@##
        for k, recomb in self._breakpoints.items():
            # NOTE: subtracting 0.5 from all recombination breakpoints,
            # to indicate that recombination 'actually' happens halfway between
            # a pair of loci, (i.e. it actually subsets Individuals'
            # genomes in that way) without having the hold all the 0.5's
            # in the Recombinations._events data struct (to save on memory)
            left = np.array([0] + [*recomb - 0.5])
            right = np.array([*recomb - 0.5] + [self._L])
            assert np.all((right-left)>0), ("Some of the following breakpoints"
                        " are invalid: :%s") % str([*zip(left, right)])
            seg_info[k] = (left, right)
        self._seg_info = seg_info


   # take a recombination event key and a parent's node ids,
   # return a zip object containing 
        # 1.) node id for parent's id (corresponding to the id in the
        # tskit.TableCollection.nodes table);
        # 2.) left end of the segment (inclusive, according to tskit)
        # 3.) right end (exclusive)
    def _get_seg_info(self, start_homologue, event_key,
                                    node_ids):
        left, right = self._seg_info[event_key]
        # add small jitter to the numbers, so that recombination breakpoints
        # are essentially not repeated, and this way tskit can correctly track
        # the number of breakpoints and number of trees
        if self._jitter_breakpoints:
            left[1:] = left[1:] + np.random.uniform(0, 0.0001, len(left)-1)
            #left, right = [np.clip(seg_bps + np.random.uniform(0, 0.0001,
            #                                                   len(seg_bps)),
            #                       0, self._L) for seg_bps in [left, right]]

            # NOTE: must make sure that each set of seg_info
            # always starts at 0 on the leftmost segment
            # (and ends at gen_arch.L on the rightmost, but that will already
            #  happen in the above code because we're only ever adding a
            #  positive value to original rightmost value, which is gen_arch.L,
            #  then clipping that value back down to gen_arch.L)
            #left[0] = 0
            right[:-1] = left[1:]

        homologue_nodes = node_ids[[(i + start_homologue) % 2 for i in range(
                                                                   len(left))]]
        seg_info = zip(homologue_nodes, left, right)
        return seg_info


class Trait:
    """
    Representation of a single trait and its parameterization.

    Trait objects are collected in a serial integer-keyed dict,
    one per trait that is parameterized for the corresponding Species,
    which inheres to the Species' GenomicArchitecture object as the attribute
    `GenomicArchitecture.traits`, from which individual Traits can be indexed
    using their index-number keys (e.g. `GenomicArchitecture.traits[<idx>]`).

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        alpha:
            A 1d numpy array containing the effect sizes of all loci subtending
            this trait. The value at index i represents the effect size of the
            locus at index i in the 'loci' attribute of this Trait.

        alpha_distr_mu:
            Mean of the normal (Gaussian) distribution
            of effect sizes for this trait's loci.

        alpha_distr_sigma:
            Standard deviation of the normal (Gaussian) distribution
            of effect sizes for this trait's loci.

        gamma:
            Curvature of the fitness function.

        idx:
            Index number of the trait (i.e. its key within the traits dict)

        loci_idxs:
            A 1d numpy array containing the index positions within an
            Individual's genotypes array
            (attribute 'g' of an Individual) that
            correspond to the loci subtending this trait.
            (This is designed such that an Individual's genotype
            for the locus number stored at Trait.loci[i]
            can be accessed by calling Individual.g[Trait.loci_idxs[i], :].)
            This method of tracking allows Individuals' full genotypes to be
            stored in tskit tables (if tskit is being used)
            while they carry their non-neutral genotypes with them
            (a computational optimization).
            Set to None if tskit is not being used.

        loci:
            A numerically sorted 1d numpy array containing
            the locus numbers of all loci subtending this trait.

        lyr_num:
            Index number of the Landscape layer to which this trait corresponds
            (i.e. of the environmental variable that serves as the selective
            force on this trait)

        max_alpha_mag:
            Maximum absolute value that can be drawn for a locus’
            effect size (i.e. alpha). Effect sizes are clipped
            to the closed interval [-max_alpha_mag, max_alpha_mag]
            when the model is parameterized.

        mu:
            Mutation rate for this trait (expressed in mutations per locus per
            time step)

        name:
            The string name of the trait

        n_loci:
            Number of loci subtending this trait

        phi:
            Polygenic selection coefficient (i.e the selection coefficient
            acting on the phenotypes, rather than the genotypes, of this Trait)

        univ_adv:
            Whether or not this trait will be treated as universally
            advantageous. When False (the default behavior), individuals will
            be fitter when their phenotypic values for this trait more closely
            align with the local environmental value of the landscape layer
            corresponding to this trait (i.e. selection will be spatially
            varying). When False, phenotypes closer to 1 will be more
            advantageous globally on the Landscape.
    """
    def __init__(self, idx, name, phi, n_loci, mu, layer, alpha_distr_mu,
                 alpha_distr_sigma, max_alpha_mag, gamma, univ_adv):
        self.idx = idx
        self.name = name
        self.phi = phi
        self.n_loci = n_loci
        if mu is None:
            mu = 0
        self.mu = mu
        self.lyr_num = layer
        self.alpha_distr_mu = alpha_distr_mu
        self.alpha_distr_sigma = alpha_distr_sigma
        self.max_alpha_mag = max_alpha_mag
        self.gamma = gamma
        self.univ_adv = univ_adv

        self.loci = np.int64([])
        self.loci_idxs = np.int64([])
        self.alpha = np.array([])

    def _get_phi(self, spp):
        if type(self.phi) in (float, int):
            phi = np.array([self.phi]*len(spp))
        else:
            phi = self.phi[spp._cells[:, 1], spp._cells[:, 0]]
        return(phi)

    def _set_loci(self, loci):
        # set the loci
        self.loci = np.hstack((self.loci, np.array([*loci])))
        self.loci.sort()
        # set the number of loci
        self.n_loci = self.loci.size

    def _set_loci_idxs(self, nonneut_loci, use_tskit):
        # if using tskit then only non-neutral loci will be stored in
        # Individuals' genotype arrays,
        # so set an index of the loci's indices in the nonneut_loci object,
        # to use when subsetting for the trait's genotypes
        if use_tskit:
            self.loci_idxs = np.array([np.where(
                                nonneut_loci == n)[0][0] for n in self.loci])
        else:
            self.loci_idxs = None

    def _add_locus(self, locus, alpha, idx):
        # get insertion index for the locus 
        insert_pt = bisect.bisect_left(self.loci, locus)
        # insert in the loci
        self.loci = np.hstack((self.loci[:insert_pt],
                               locus,
                               self.loci[insert_pt:]))

        # insert the effect size for the locus
        self.alpha= np.hstack((self.alpha[:insert_pt],
                               alpha,
                               self.alpha[insert_pt:]))

        # insert in the loci_idxs
        # NOTE: add 1 to all superseding locus indexes, to account for
        # locus newly inserted into genotype arrays
        self.loci_idxs = np.hstack((self.loci_idxs[:insert_pt],
                                  idx,
                                  self.loci_idxs[insert_pt:] + 1))

        # increment the number of loci
        self.n_loci += 1


class GenomicArchitecture:
    """
    Representation of the genomic architecture of a Species.

    Inheres to its Species as the Species.gen_arch attribute.

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        delet_alpha_distr_scale:
            The scale parameter of the Gamma distribution from which
            the effect sizes (i.e. alphas) of deleterious loci are drawn

        delet_alpha_distr_shape:
            The shape parameter of the Gamma distribution from which
            the effect sizes (i.e. alphas) of deleterious loci are drawn

        delet_loci:
            A 1d numpy array that tracks all deleterious loci

        delet_loci_idxs:
            A 1d numpy array containing the index positions within an
            Individual's genotypes array (attribute 'g' of an Individual)
            that correspond to the deleterious loci.
            (This is designed such that an Individual's genotype for
            the deleterious locus number stored at
            GenomicArchitecture.delet_loci[i] can be accessed by calling
            Individual.g[GenomicArchitecture.delet_loci_idxs[i], :].)
            This method of tracking allows Individuals' full genotypes to be
            stored in the tskit tables (if tskit is being used)
            while they carry their non-neutral genotypes with them
            (a computational optimization).
            Set to None if tskit is not being used.

        delet_loci_s:
            A 1d numpy array containing the selection strengths
            of each of the deleterious loci. (Selection strengths are
            listed in an order corresponding to the order of
            GenomicArchitecture.delet_loci, such that the selection strength of
            the deleterious locus referenced at
            GenomicArchitecture.delet_loci[i] is stored at
            GenomicArchitetcture.delet_loci_s[i].)

        dom:
            Dominance values for all loci (stored as an L-length,
            1d numpy array in which 0 indicates a codominant locus
            and 1 indicates a locus for which the '1' allele is dominant)

        L:
            Length of the simulated genome

        mu_delet:
            Genome-wide deleterious mutation rate (expressed in mutations
            per locus per time step). Deleterious mutations do not influence
            trait phenotypes, but instead are univerally
            deleterious (to an extent determined by their effect sizes).

        mu_neut:
            Genome-wide neutral mutation rate (expressed in mutations
            per locus per time step).

        neut_loci:
            A 1d numpy array that tracks all loci that do not influence the
            phenotypes of any traits.

        nonneut_loci:
            A 1d numpy array that tracks all loci that either influence the
            phenotype of at least one trait or are deleterious.

        p:
            Starting allele frequencies for all loci (stored as an L-length,
            1d numpy array)

        pleiotropy:
            A bool indicating whether or not to allow pleiotropy (i.e. whether
            or not to allow the same locus to subtend multiple traits)

        recombinations:
            The Recombinations object, which contains the parameters and data
            necessary for simulation of recombination. Not intended for public
            use at this time.

        sex:
            A bool indicating whether or not Individuals of this Species
            will be assigned a sex (0 for female, 1 for male)

        traits:
            A dict containing all of a Species' Trait objects, keyed by serial
            integers.

        x:
            Ploidy
            NOTE: only diploidy is currently implemented, but this attribute
                  serves as a placeholder for use in possible
                  future generalization to arbitrary ploidy.)
    """
    def __init__(self, dom, g_params, land, recomb_rates=None,
                 recomb_positions=None):
        # ploidy (NOTE: for now will be 2 by default; later could consider
        # enabling polyploidy)
        self.x = 2
        # total length (i.e. number of markers)
        self.L = g_params.L
        # placeholder for the starting allele freqs for all loci
        self.p = None
        # True/False regarding whether to allow a locus to affect the
        # phenotype of more than one trait; defaults to False
        self.pleiotropy = g_params.pleiotropy
        # array of dominance values for all loci
        self.dom = dom
        # set the _use_dom attribute based on whether any loci have a 1 for
        # their dominance value
        self._use_dom = np.any(self.dom)
        # whether or not to use sexes for this species
        self.sex = g_params.sex

        # whether or not to use tskit to track this species' spatial pedigree
        self.use_tskit = g_params.use_tskit
        assert self.use_tskit in [True, False], (("'use_tskit' must be either "
                                                  "True or False for each "
                                                  "Species."))
        # set interval for simplifying the tskit TableCollection
        self.tskit_simp_interval = g_params.tskit_simp_interval
        if self.use_tskit:
            assert isinstance(self.tskit_simp_interval,
                              int), (("If using tskit then the "
                                      "'tskit_simp_interval' parameter must be "
                                      " an integer."))

        # genome-wide neutral mutation rate
        self.mu_neut = g_params.mu_neut
        # array to keep track of all loci that don't influence the
        # phenotypes of any trait; defaults to all loci, then will be updated
        self.neut_loci = np.array(range(self.L))
        # array to keep track of all loci that influence the phenotype of at
        # least one trait; after burn-in, will be updated as needed
        self.nonneut_loci = np.array([])

        # genome-wide deleterious mutation rate
        self.mu_delet = g_params.mu_delet
        self.delet_alpha_distr_shape = g_params.delet_alpha_distr_shape
        self.delet_alpha_distr_scale = g_params.delet_alpha_distr_scale


        # arrays to track deleterious loci, their genotype indices, and their
        # strengths of selection
        self.delet_loci = np.int64([])
        self.delet_loci_idxs = np.int64([])
        self.delet_loci_s = np.array([])

        # add a dict of Trait objects, if necessary
        self.traits = None
        if 'traits' in [*g_params]:
            self.traits = _make_traits(g_params.traits, land)

        # set self._mu_tot, the total per-site, per-generation mutation rate
        mus = [mu for mu in (self.mu_neut, self.mu_delet) if mu is not None]
        if self.traits is not None:
            mus = mus + [trt.mu for trt in self.traits.values()]
        self._mu_tot = sum(mus)
        # save the nonneutral mutation rate
        self._mu_nonneut = self._mu_tot - self.mu_neut

        # set a placeholder for the species' mutable loci
        # (to be filled after burn-in)
        self._mutables = None

        # attribute that will be replaced with the estimated total number of
        # mutations per iteration, once it's estimated at the end of the first
        # burn-in
        self._mut_fns = self._make_mut_fns_dict()
        # set ._planned_muts to None, for now (this is not yet implemented,
        # but thinking about it
        self._planned_muts = None

        # The recombination-paths object will be assigned here; used to
        # speed up large quantities of binomial draws needed for recombination
        self.recombinations = Recombinations(self.L, recomb_positions,
                                             g_params.n_recomb_sims,
                                             g_params.r_distr_alpha,
                                             g_params.r_distr_beta,
                                             recomb_rates,
                                             g_params.jitter_breakpoints)

    # method to make a _mut_fns dict, containing a function
    # for each type of mutation for this species
    def _make_mut_fns_dict(self):
        mut_fns = {}
        if self.mu_neut > 0:
            def neut_fn(spp, offspring):
                return(mutation._do_neutral_mutation(spp, offspring))
            mut_fns.update({'neut': neut_fn})
        if self.mu_delet > 0:
            def delet_fn(spp, offspring):
                return(mutation._do_deleterious_mutation(spp, offspring))
            mut_fns.update({'delet': delet_fn})
        if self.traits is not None:
            for trait_num in self.traits:
                if self.traits[trait_num].mu > 0:
                    def trait_fn(spp, offspring, trait_num=trait_num):
                        return(mutation._do_trait_mutation(spp, offspring,
                                                           [trait_num]))
                    mut_fns.update({'t%i' % trait_num: trait_fn})
        return mut_fns

    # method to draw mutation types for any number of mutations chosen
    # to occur in a given timestep
    def _draw_mut_types(self, num):
        type_dict = {'neut': self.mu_neut,
                     'delet': self.mu_delet}
        if self.traits is not None:
            trait_dict = {'t%i' % (k): v.mu for k, v in self.traits.items()}
            type_dict.update(trait_dict)
        types = []
        probs = []
        for k, v in type_dict.items():
            types.append(k)
            probs.append(v)
        probs = [p/sum(probs) for p in probs]
        choices = r.choice(types, p=probs, size=num, replace=True)
        return(choices)

    # method for drawing an effect size for one or many loci
    def _draw_trait_alpha(self, trait_num, n=1):
        mu = self.traits[trait_num].alpha_distr_mu
        sigma = self.traits[trait_num].alpha_distr_sigma
        max_alpha_mag = self.traits[trait_num].max_alpha_mag
        if max_alpha_mag is not None:
            min_alpha = -1 * max_alpha_mag
            max_alpha = max_alpha_mag
        else:
            min_alpha = max_alpha = max_alpha_mag
        # use mu value as the fixed effect size, if sigma is 0
        if sigma == 0:
            alpha = mu * np.array([1 - (i % 2)*2 for i in range(n)])
        else:
            alpha = r.normal(self.traits[trait_num].alpha_distr_mu,
                             self.traits[trait_num].alpha_distr_sigma, n)
            if max_alpha_mag is not None:
                alpha = np.clip(alpha, min_alpha, max_alpha)
        # otherwise use mu and sigma to draw effects from
        # (because effects will be added to 0)
        if self.traits[trait_num].n_loci == 1:
            alpha = np.abs(alpha)
        return(alpha)

    # method for drawing new deleterious mutational fitness effects
    def _draw_delet_s(self):
        s = r.gamma(self.delet_alpha_distr_shape, self.delet_alpha_distr_scale)
        s = min(s, 1)
        return(s)

    # method for assigning loci to traits
    def _set_trait_loci(self, trait_num, mutational=False,
                        loci=None, alpha=None):
        # if this is not the result of a point mutation, but instead
        # either an initial setup or manually introduced, then grab the
        # number of loci to be assigned
        if not mutational:
            n = self.traits[trait_num].n_loci
            assert n <= self.L, ("The number of loci parameterized for "
                                 "trait number %i ('n_loci') is greater "
                                 "than the length of the genome!")
        # otherwise, assign a single locus
        else:
            n = 1
        if loci is not None:
            if not np.iterable(loci):
                loci = [loci]
            assert len(set([*loci])) == len(loci), ("Some of the trait "
                                                    "loci provided appear "
                                                    "to be repeated.")
        # else, draw loci randomly, either allowing pleiotropy or not
        elif not self.pleiotropy:
            loci = set(r.choice(self.neut_loci, size=n, replace=False))
        elif self.pleiotropy:
            loci = set(r.choice(range(self.L), size=n, replace=False))
        # update the trait's loci
        self.traits[trait_num]._set_loci(loci)
        # add these loci to self.non-neutral and remove from
        # self.neut_loci, to keep track of all loci underlying any
        # traits (for purpose of avoiding pleiotropy)
        self.nonneut_loci = np.array(sorted([*self.nonneut_loci] + [*loci]))
        self.neut_loci = np.array(sorted([*set(self.neut_loci).difference(
                                                    set(self.nonneut_loci))]))
        # if the effect size(s) is/are provided, use those
        if alpha is not None:
            if not np.iterable(alpha):
                alpha = np.array([alpha])
            effects = np.array([*alpha])
        # else, draw effects from a Gaussian dist with mean 0 and sigma
        # provided by trait params (as per e.g. Yeaman and Whitlock 2011)
        else:
            effects = self._draw_trait_alpha(trait_num, n)
        # if this is drawing the effect of a monogenic trait's locus (i.e. if
        # it's drawing only one value, but it's not mutational), then coerce
        # the effects to [0.5]
        if not mutational and n == 1:
            effects = np.array([0.5])
        # check that loci and effects are of equal length
        assert len(loci) == len(effects), ('Lengths of the two arrays '
                                           'containing the new trait loci and '
                                           'their effects are not equal.')
        # then add the effects to the trait's alpha array
        self.traits[trait_num].alpha = np.hstack((self.traits[trait_num].alpha,
                                                  effects))

    # add a nonneutral locus to the genomic architecture
    # NOTE: either trait_nums or delet_s must be non-None,
    # and trait_nums must be iterable (if not None)
    def _add_nonneut_locus(self, locus, trait_nums=None, delet_s=None):
        # remove from the neut_loci array
        self.neut_loci = np.delete(self.neut_loci,
                                   np.where(self.neut_loci == locus))
        # add the locus to the nonneut_loci array
        idx = bisect.bisect_left(self.nonneut_loci, locus)
        self.nonneut_loci = np.hstack((self.nonneut_loci[:idx],
                                       locus,
                                       self.nonneut_loci[idx:]))

        # add the locus to either trait loci, if necessary
        if trait_nums is not None and delet_s is None:
            for n in trait_nums:
                alpha = self._draw_trait_alpha(n)[0]
                self.traits[n]._add_locus(locus, alpha, idx)
        # or else add to the deleterious loci
        elif delet_s is not None and trait_nums is None:
            # TODO: DECIDE IF NEED TO USE self.loci INSTEAD OF self.delet_loci
            #       HERE IF use_tskit==False
            del_idx = bisect.bisect_left(self.delet_loci, locus)
            # add the locus, its genome-index (for subsetting individuals'
            # genomes when calculating fitness), and its strength of selection
            # to the deleterious locus trackers
            self.delet_loci = np.hstack((self.delet_loci[:del_idx],
                                         locus,
                                         self.delet_loci[del_idx:]))
            if self.use_tskit:
                self.delet_loci_idxs = np.hstack(
                    (self.delet_loci_idxs[:del_idx], idx,
                     self.delet_loci_idxs[del_idx:]))
            else:
                self.delet_loci_idxs = None
            self.delet_loci_s = np.hstack((self.delet_loci_s[:del_idx],
                                         delet_s,
                                         self.delet_loci_s[del_idx:]))
        return idx


    # method for plotting all allele frequencies for the species
    def _plot_allele_frequencies(self, spp, color='red'):
        if spp.gen_arch.use_tskit:
            speciome = spp._get_genotypes(biallelic=False)
        else:
            speciome = np.mean(np.stack([ind.g for ind in spp.values()]), axis=2)
        freqs = np.mean(speciome, axis=0)
        assert len(freqs) == spp.gen_arch.L
        plt.plot(range(self.L), self.p, ':k', label='start freq.', alpha=0.5)
        plt.plot(range(self.L), freqs, '-', color=color, label='curr. freq.')
        plt.xlabel('locus')
        plt.ylabel('frequency')
        plt.legend()
        plt.show()

    # method for pickling a genomic architecture
    def _write_pickle(self, filename):
        import cPickle
        with open(filename, 'wb') as f:
            cPickle.dump(self, f)


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

# generate allele_freqs
def _draw_allele_freqs(l):
    return(r.beta(1, 1, l))


def _make_traits(traits_params, land):
    params_copy = {**traits_params}
    # and set each Layer number using Layer names
    for k, v in params_copy.items():
        # the first time this is run during a model with random
        # communities at each iteration, the layer identified in
        # the traits_params will be a string indicating the layer's name
        if isinstance(v.layer, str):
            lyr_num = [num for num, lyr in land.items(
                                                ) if lyr.name == v.layer]
        # the second and later times it's run during such a model (see
        # previous comment), it will already have been swapped out for an int
        # indicating the layer's index num
        elif isinstance(v.layer, int):
            lyr_num = [num for num, lyr in land.items(
                                                ) if lyr.idx == v.layer]
        assert len(lyr_num) == 1, ("Expected to find a single Layer with "
                                   "the Layer name indicated for Trait %s, "
                                   "but instead found "
                                   "%i.") % (k, len(lyr_num))
        v['layer'] = lyr_num[0]
    # then for each of i traits, unpack the ith components of the remaining
    # params to create the trait dict
    traits = {n: Trait(n, k_v[0], **k_v[1]) for n, k_v in enumerate(
                                                        params_copy.items())}
    # for all monogenic traits, if the trait doesn't already have a 0 mutation
    # rate then coerce it to 0
    # NOTE: this is the because there is a good reason to use 0 as the 
    # baseline phenotype for monogenic traits but 0.5 as the baseline
    # for multilocus traits (to preserve symmetry of ability for individuals
    # to have phenotypes beyond 0 and 1 for a multilocus trait and thus
    # experience stabilizing selection within their optimal habitat), but
    # this means that there would be a huge problem if a monogenic trait
    # underwent an adaptive mutation and became polyenic bceause all of the
    # individuals' phenotypes would suddenly have to be recalculated and would
    # suddenly completely change
    for n, trt in traits.items():
        if trt.n_loci == 1:
            if trt.mu != 0:
                warnings.warn(("Coercing Trait %i ('%s') to a "
                               "0 mutation rate because it is monogenic."))
                trt.mu = 0
    return(traits)


# build the genomic architecture
def _make_genomic_architecture(spp_params, land):
    # get the genome parameters
    g_params = spp_params.gen_arch
    # get the custom genomic-architecture file, if provided
    gen_arch_file = None
    if 'gen_arch_file' in g_params.keys():
        if g_params.gen_arch_file is not None:
            gen_arch_file = pd.read_csv(g_params.gen_arch_file)
            # check that the genome length matches the params file
            assert len(gen_arch_file) == g_params.L, ("The length of the "
                        "custom genomic architecture file must match the "
                        "genome length provided to the parameter 'L' in "
                        "Geonomics parameters file.")
            # check that each trait is assigned an alpha value
            for trt_val, alpha_val in zip(gen_arch_file['trait'],
                                          gen_arch_file['alpha']):
                if pd.notnull(trt_val):
                    assert pd.notnull(alpha_val), ("All trait-associated loci "
                                                   "specified in a custom "
                                                   "genomic architecture file "
                                                   "must be assigned non-null "
                                                   "alpha values. Please check "
                                                   "your file, then rerun.")

            # ensure that trait and alpha columns are strings to start (because
            # in many cases alpha would likely read in as a float, if each row
            # has at most one alpha value because there's no pleiotropy, but in
            # case of pleiotropy I need to be able to use the str.split method
            # on each row's values further down)
            gen_arch_file['trait'] = [str(v) for v in gen_arch_file['trait']]
            gen_arch_file['alpha'] = [str(v) for v in gen_arch_file['alpha']]
            assert (np.all(
                (gen_arch_file['dom'] == 0) + gen_arch_file['dom'] == 1)), (
                "The 'dom' column of the custom genomic architecture file "
                "must contain only 0s and 1s (where 0 indicates codominance "
                "of the 0 and 1 alleles, 1 indicates that the 1 allele "
                "is dominant).")
            # check that each trait has in that file the number of loci
            # indicated by n_loci in that trait's params dict
            if 'traits' in [*g_params]:
                all_traits = [trt.strip() for row in [val.split(
                    ',') for val in gen_arch_file['trait']] for trt in row]
                for trt_name, trt in g_params.traits.items():
                    n_loci_in_file = sum(
                                    [trt == trt_name for trt in all_traits])
                    assert_msg = ("The number of times a Trait is appears "
                                  "in the custom genomic architecture file "
                                  "must be equivalent to the number of loci "
                                  "subtending that Trait as indicated by the "
                                  "'n_loci' key in its section of the "
                                  "parameters file.")
                    assert n_loci_in_file == trt.n_loci, assert_msg

    # get the sex parameter and add it as an item in g_params
    g_params['sex'] = spp_params.mating.sex

    # get the custom recomb_rates and recomb_positions from the
    # custom gen-arch file, if provided
    # NOTE: each recombination rate
    #       stipulated in gen_arch file
    #       is construed as the recombination rate
    #       at the midway point between that locus
    #       and the previous one
    #       (NOTE: this is of course non-sensical for the 0th
    #              locus, which would thus have its rate expressed as
    #              the rate of recomb between the -1th and the 0th,
    #              but that's fine because the first rate is later coerced to 0
    #              anyhow, then the starting homologue of the recomb path,
    #              i.e. the sister chromatid placed into the gamete,
    #              is later chosen using a ~Bern(0.5) draw at the time
    #              mating occurs; see mating.py)
    recomb_rates = None
    recomb_positions = None
    if gen_arch_file is not None:
        recomb_rates = gen_arch_file['r'].values
        recomb_positions = gen_arch_file['locus'].values

    # set locus-wise dominance values for the 1-alleles, using the 'dom' value
    # in the gen_arch params, unless a gen_arch_file was provided
    if gen_arch_file is None:
        # create an L-length array of boolean integers (where 0 = codominance,
        # 1 = dominance)
        dom = np.array([int(g_params.dom)] * g_params.L)
    else:
        # get the 'dom' column from the gen_arch_file
        dom = gen_arch_file['dom'].values

    # now make the gen_arch object
    gen_arch = GenomicArchitecture(dom, g_params, land,
                                   recomb_rates, recomb_positions)

    # set the loci and effect sizes for each trait, using the custom gen-arch
    # file, if provided, and if the species has traits
    if gen_arch_file is not None and gen_arch.traits is not None:
        # convert the trait names in the 'trait' column of the file into
        # lists of their trait numbers (i.e. their keys in the gen_arch
        # traits dict)
        trt_names_nums = {
            trt.name: num for num, trt in gen_arch.traits.items()}
        gen_arch_file['trait'] = [[trt_names_nums[
            val] for val in [x.strip() for x in row.split(
            ',')] if val in trt_names_nums] for row in gen_arch_file['trait']]
        # turn the values in the 'alpha' column into lists of
        # values, by splitting on commas
        # (this will allow people to assign a single locus
        # to more than one trait, i.e. to model pleiotropy)
        gen_arch_file['alpha'] = [
            [*map(float, row.split(','))] for row in gen_arch_file['alpha']]
        # get the loci and effect sizes for each trait
        loci = {}
        alphas = {}
        for trt_num in gen_arch.traits.keys():
            loci[trt_num] = np.array([*gen_arch_file['locus']])[
                [trt_num in row for row in gen_arch_file['trait']]]
            alphas[trt_num] = np.concatenate([np.array(row['alpha'])[
                [n == trt_num for n in row[
                    'trait']]] for i, row in gen_arch_file.iterrows()])
        # check that we got the same length of loci and effect sizes for
        # a trait, for all traits
        for trt_num in loci.keys():
            assert_msg = ("Expected to receive the same number of loci and "
                          "alphas (i.e. effect sizes) for trait number %i, "
                          "but instead got %i loci and %i "
                          "alphas.") % (trt_num, len(loci[trt_num]),
                                        len(alphas[trt_num]))

            assert len(loci[trt_num]) == len(alphas[trt_num]), assert_msg

        # add the loci and effect sizes for each of the traits to the
        # Trait object in the GenomicArchitecture
        for trait_num in gen_arch.traits.keys():
            gen_arch._set_trait_loci(trait_num, mutational=False,
                                     loci=loci[trait_num],
                                     alpha=alphas[trait_num])
    # or else randomly set the loci and effect sizes for each trait
    else:
        if gen_arch.traits is not None:
            for trait_num in gen_arch.traits.keys():
                gen_arch._set_trait_loci(trait_num, mutational=False)

    # now that all nonneutral loci have been selected, 
    # set each trait's loci_idxs (which maps a trait's numeric loci to
    # their index positions in individuals' genomes)
    # NOTE: will just be set to None if not using tskit
    if gen_arch.traits is not None:
        for trt in gen_arch.traits.values():
            trt._set_loci_idxs(gen_arch.nonneut_loci, gen_arch.use_tskit)

    assert_msg = ("The union of the gen_arch.neut_loci and "
                  "gen_arch.nonneut_loci sets does not contain all loci "
                  "indicated by gen_arch.L")
    assert len(set(range(gen_arch.L)).difference(
        set(gen_arch.neut_loci).union(
        set(gen_arch.nonneut_loci)))) == 0, assert_msg

    # draw locus-wise 1-allele frequencies, unless provided in
    # custom gen-arch file
    if gen_arch_file is None:
        # set according to a non-None value
        if g_params.start_p_fixed is not None:
            if isinstance(g_params.start_p_fixed, bool):
                # if True, make all 0.5 (default starting allele freq)
                if g_params.start_p_fixed:
                    gen_arch.p = np.array([0.5]*g_params.L)
                # if False, draw randomly
                else:
                    gen_arch.p = _draw_allele_freqs(g_params.L)
            # if not a bool then should be a float, so assert,
            # then set to that val
            else:
                assert 0 <= g_params.start_p_fixed <= 1, ("If a starting allele"
                                                        " frequency value is "
                                                        "provided then it must "
                                                        "be between 0 and 1.")
                gen_arch.p = np.array([g_params.start_p_fixed]*g_params.L)
        # if the value is None, then draw randomly
        else:
            gen_arch.p = _draw_allele_freqs(g_params.L)
        # set the neutral loci to 0, if need be
        if g_params.start_neut_zero and len(gen_arch.neut_loci) > 0:
            gen_arch.p[gen_arch.neut_loci] = 0
    # use starting allele freq values from the gen_arch_file instead, if given
    else:
        gen_arch.p = gen_arch_file['p'].values

    # set the recombination events
    use_subsetters = ((gen_arch.use_tskit and (len(gen_arch.nonneut_loci) > 0 or
                                               gen_arch._mu_nonneut > 0))
                      or not gen_arch.use_tskit)
    gen_arch.recombinations._set_events(use_subsetters, gen_arch.nonneut_loci,
                                        gen_arch.use_tskit)

    return gen_arch


# reset genomes after burn-in
def _check_mutation_rates(gen_arch, est_tot_muts, burn_T, T):
    # if there do not appear to be enough space in the simulated genome for
    # the expected number of mutations then warn the user of that
    if est_tot_muts > 0.75 * len(
        [loc for loc in range(
                    gen_arch.L) if loc not in gen_arch.nonneut_loci]):
        raise MutationRateError(("This species has been "
                                 "parameterized with too few neutral "
                                 "loci to accommodate the expected "
                                 "number of mutations. (Geonomics only "
                                 "uses an infinite sites model.) "
                                 "Please tweak some combination of "
                                 "the genome length, model run time, "
                                 "or mutation rates."))


    # skip this step and force the neutral mutation rate to 0, if there are no
    # neutral loci in the genome as it was configured
    if len(gen_arch.neut_loci) == 0 and gen_arch._mu_tot > 0:
        warn_msg = ("This species has been parameterized "
                    "with non-zero mutation rates but "
                    "without any neutral loci, leaving no target "
                    "for mutations. Please tweak the genome "
                    "length and/or mutation rates.")
        warnings.warn(warn_msg)
        gen_arch.mu_neut = 0
        gen_arch.mu_delet = 0
        for trt in gen_arch.traits.values():
            trt.mu = 0
    # or just outright skip this step if the parameterization included only
    # mutation rates of 0
    elif gen_arch._mu_tot == 0:
        pass
    # otherwise, set the mutable loci
    else:
        mutables = [*set(range(gen_arch.L)).difference(
                                            set(gen_arch.nonneut_loci))]
        r.shuffle(mutables)
        gen_arch._mutables = [*mutables]
    return


def _make_starting_mutations(spp, tables):
    '''
    function to generate mutations, after burn-in,
    and to assign them to a species' TableCollection's current nodes,
    to produce the starting 1-allele frequencies parameterized for the species
    '''
    # get the starting frequencies for each site
    start_freqs = spp.gen_arch.p

    # get a set of all homologues, as tuples of (individual id, homologue idx)
    homologues = [*zip(np.repeat([*spp], 2),
                         [*range(spp.gen_arch.x)] * len(spp))]

    # make mutations for each site
    for site, freq in enumerate(start_freqs):
        # generate the number of mutations for this locus
        n_mutations = int(round(2 * len(spp) * freq, 0))
        # make sure we don't mutate either all or none of the population's
        # homologues, unless called for
        if n_mutations == len(spp) * 2 and freq < 1:
            n_mutations -= 1
        if n_mutations == 0 and freq > 0:
            n_mutations = 1
        #randomly choose and mutate n_mutations homologues from the population 
        np.random.shuffle(homologues)
        homologues_to_mutate = homologues[:n_mutations]
        for ind, homol in homologues_to_mutate:
            # if using tskit, add muts to individis' genotype arrays as needed,
            # and also to mutations table
            if spp.gen_arch.use_tskit:
                # create a mutation in the individual's genome, if this is a
                # non-neutral locus
                if site in spp.gen_arch.nonneut_loci:
                    spp[ind].g[np.where(spp.gen_arch.nonneut_loci == site),
                               homol] = 1
                # get the homologue's nodes-table id,
                # then add a row to the mutations table
                node_id = spp[ind]._nodes_tab_ids[homol]
                tables.mutations.add_row(site, node=node_id, parent=-1,
                                         derived_state='1')

            # if not using tskit just add all muts to individs' genotypes arrays
            else:
                spp[ind].g[site, homol] = 1

    # and then reset the individuals' phenotypes, if needed
    if spp.gen_arch.traits is not None:
        [ind._set_z(spp.gen_arch) for ind in spp.values()]

    return


# method for loading a pickled genomic architecture
def read_pickled_genomic_architecture(filename):
    import cPickle
    with open(filename, 'rb') as f:
        gen_arch = cPickle.load(f)
    return gen_arch


################################################################
# FUNCTIONS FOR WORKING WITH msprime TO SIMULATE GNX INDIVIDUALS
################################################################

def _format_msprime_rate(rate, L):
    """
    Either return a simple float, if the given rate
    is homogeneous, or else return an msprime.RateMap object.
    Used for msprime recombination and mutation rates.
    """
    assert (type(rate) in [float, list, np.ndarray] or rate == 0 or rate == 1)
    if (isinstance(rate, float) or rate == 0 or rate == 1):
        return rate
    elif len(np.unique(rate)) == 1:
        return rate[0]
    else:
        assert len(rate) == L
        # NOTE: position array has to be one longer than the sequence length,
        #       so that rightmost position brackets a stretch of genomic
        #       intervals equal to the sequence_length
        positions = np.linspace(0, L, L+1)
        ratemap = msprime.RateMap(position=positions, rate=rate)
        return ratemap


def _sim_msprime_individuals(n,
                             L,
                             recomb_rate,
                             mut_rate,
                             demography=None,
                             population_size=None,
                             ancestry_model=None,
                             random_seed=None,
                             use_tskit=True,
                            ):
    '''
    Use an msprime simulation to generate new Geonomics Individuals.
    (docs in the main.py wrapper)
    '''
    # handle the recombination and mutation rates
    if (not isinstance(recomb_rate, float) and
        not (recomb_rate == 0)):
        # NOTE: tack a 0.5 onto the righthand side of the recomb_rate vector,
        #       to satisfy the difference between msprime (for which L
        #       expresses length in intervals, each of which needs a recomb
        #       rate specified, hence the recomb_rate vector needing to have a
        #       number of rates equal to L) and gnx (for which L expressed
        #       length in loci, between which recomb rates are needed, hence
        #       needing L-1 recomb rates);
        #       TODO: DOUBLE-CHECK THE FOLLOWING LOGIC
        #       this 0.5 value will not affect gnx data because it effectively
        #       sits between sites L-1 and L, but a simulated gnx genome of
        #       length L only includes sites [0, L-1]
        recomb_rate = [*recomb_rate] + [0.5]
    recomb_rate = _format_msprime_rate(recomb_rate, L)
    mut_rate = _format_msprime_rate(mut_rate, L)

    # simulate ancestry and mutations
    # TODO: look back at how gnx uses genome positions, because 
    #       it looks like we need to provide discrete_genome=false
    #       to sim_ancestry and true to sim_mutations;
    #       or perhaps instead just fine to use false for both
    #       and then subtract 0.5 from all non-zero 'left' values
    #       from all non-L 'right' values when handles the edges table?
    ts = msprime.sim_ancestry(samples=n,
                              sequence_length=L,
                              demography=demography,
                              population_size=population_size,
                              recombination_rate=recomb_rate,
                              ploidy=2,
                              model=ancestry_model,
                              random_seed=random_seed,
                              discrete_genome=True,
                             )
    # NOTE: mutations use a simple binary 0|1 model, with 0 always the
    #       ancestral state, and with mutations being state-independent
    #       (i.e., always placing a 1, even if this results in a silent
    #       mutation of 1→1);
    #       this model is the best aligned to gnx data
    mut_model = msprime.BinaryMutationModel(state_independent=True)
    mts = msprime.sim_mutations(tree_sequence=ts,
                                rate=mut_rate,
                                model=mut_model,
                               )
    # get the mutated tskit.TableCollection object
    mtc = mts.dump_tables()

    # double-check the number of Individuals
    assert np.sum(mtc.nodes.asdict()['individual']>-1) == 2*n

    # create the Individuals, assigning placeholder values for the attributes
    # (to be later updated by downstream code)
    individs = [Individual(idx=0, x=0, y=0) for i in range(n)]
    # get the information linking Individuals to rows in TableCollection tables
    nodes_tab_individs_col = mtc.nodes.asdict()['individual']
    for i, ind in enumerate(individs):
        # get the nodes table rows corresponding to this Individual
        nodes_tab_ids = np.where(nodes_tab_individs_col==i)[0]
        assert len(nodes_tab_ids) == 2
        # set hidden attributes tracking the individs' rows in the 'nodes' table
        # and id in the 'individuals' table
        ind._set_nodes_tab_ids(nodes_tab_ids)
        ind._individuals_tab_id = i
        # if the gnx Model these Individuals are intended for is not using tskit
        # then walk the variants generator and compile the individ's
        # genotype array, then store it in the attribute 'g'
        if not use_tskit:
            var_gen = mts.variants(samples=nodes_tab_ids)
            gt = np.stack([next(var_gen).genotypes for _ in range(L)])
            ind.g = gt

    return individs, mtc


###############################################################################
# FUNCTIONS FOR WORKING WITH THE SPATIAL PEDGREE SAVED IN tskit DATA STRUCTURES
###############################################################################

def _prep_tskit_tabcoll_for_gnx_spp_union(tc,
                                          recip_spp,
                                          coords,
                                          source_software,
                                         ):
    '''
    Reformat a tskit.TableCollection, produced by either a gnx simulation
    or an msprime simulation, so that it conforms to Geonomics' expected format
    and has the other data it needs to be fed into the .union() method
    of the given Geonomics Species' current TableCollection.

    The code and comments walk through each table, column by column,
    and specify if and how the contents need to be reformatted or updated
    in order to match Geonomics conventions.

    Returns the list of gnx Individual idx numbers that were set in the
    TableCollection, for downstream use and/or validation.
    '''
    # check source_software is valid
    assert source_software in ['gnx', 'msprime'], ('This function can only '
                                                   'prep TableCollections '
                                                   'that originate from '
                                                   'gnx Models or msprime '
                                                   'simulations.')

    # use the coords array to get the number of new individuals being unioned
    # into the recipient population
    assert (isinstance(coords, np.ndarray) and
            len(coords.shape)==2 and
            coords.shape[1] == 2), ("The coords param must be a 2d "
                                    "numpy.ndarray with the x,y coords in "
                                    "the 2 columns.")
    n_source_individs = coords.shape[0]
    if source_software == 'msprime':
        assert tc.individuals.num_rows == n_source_individs

    # reformat nodes table content:
    #-----------------------------

    if source_software == 'msprime':
        # gnx uses the time column as follows:
            # - smallest values are the most recent individuals
            # - nodes with val 1 (i.e. -1 * t=-1 before main phase begins) are
            #   nodes that originated at the end of the burn-in phase
            # - nodes with val 0, -1, -2, ... (i.e., -1 * t =0, 1, 2,...)
            #   originated during those time steps
        # msprime times are expressed as t before present (i.e. samples are 0s).
        # thus, convert msprime to gnx convention by subtracting from the whole
        # column the value of spp.t at the time of their introduction
        # (i.e., the time of their simulation and sampling)
        time = tc.nodes.time - recip_spp.t
        assert np.all([t>=(-1*recip_spp.t) for t in time]), ("Values "
                            "attempting to be set on 'time' column of nodes "
                            "table are invalid (i.e., are more negative than "
                            "-1 * recipient Species' current model time, and "
                            "thus would refer to future time steps not yet "
                            "simulated.")
        tc.nodes.time = time

        # msprime uses the flags column to indicate extant nodes
        # (i.e., docs say "NODE_IS_SAMPLE = 1"); gnx uses flags to distinguish
        # between nodes in 'real' gnx Individuals and msprime nodes;
        # thus, we actually just want to keep this column as is, because the
        # samples are only nodes that will be assigned to new gnx Individuals,
        # and thus are the only ones we would otherwise coerce to 1s anyhow!

        # nothing to do the population column; TableCollection.union()
        # will be asked to use the next available int in the recipient Species'
        # populations table to automatically record the
        # population these nodes came from

        # the individual column needn't be touched; TableCollection.union()
        # will automatically increment this col by using the next individual id
        # available in the recipient Species' individuals table;
        # we will just need to later track what the next id was and how many
        # individuals were added, so that TableCollection.individuals ids can
        # be matched to gnx Individuals via Individuals._individuals_tab_id attr

        # metadata, metadata_offset, and metadata_schema columns are currently
        # unused by both msprime and gnx

    elif source_software == 'gnx':
        # no need to alter any of this data if it's already from a gnx model
        pass



    # reformat individuals table content:
    #-----------------------------------

    if source_software == 'msprime':
        # all individuals in the individuals table are extant and will be turned
        # into gnx Individual objects, and gnx uses a 1 in the 'flags' column to
        # indicated 'real' Individuals (i.e., individuals that were actual
        # gnx.Individuals at some point), so set the entire 'flags' column to 1s
        flags = tc.individuals.flags*0+1
    elif source_software == 'gnx':
        # flags should already be correct for a gnx model,
        flags = tc.individuals.flags

    # gnx stores individs' starting x,y coordinates in the 'location' column
    # (NOTE: also the phenotype and fitness values, but those are not relevant
    # in the neutral-loci-only msprime simulations permitted by gnx);
    # TODO: think about if/how sweeps would be allowed and
    #       how they would be represented/matched up to genarch
    # location_offset tells the TableCollection to jump 2 values (x, y) per row
    # NOTE: 'ragged columns' actually consist of the column's contents in a
    #       flattened array format and a column of n+1 index values that can be
    #       offset and used to index the start and end positions of each row's
    #       contents within that array
    if source_software == 'msprime':
        location_offset = [int(n) for n in np.linspace(0,
                                                   2*tc.individuals.num_rows,
                                                   tc.individuals.num_rows+1)]
    elif source_software == 'gnx':
        # in a TableCollection from a gnx Model, locations will need to be
        # blanked out for all Individuals, current or past (so that fns that
        # later depend on the stored locations information don't accidentally
        # use data conveying locations on a different landscape from before the
        # introduction of these source Individuals to their new population),
        # and then will need to be reset to the new coords values for only the
        # bottom n rows; thus, need to tack leading zeros onto the
        # location_offset values just generated, making the final vector
        # num_rows+1 long
        location_offset = [int(n) for n in np.linspace(0,
                                                   2*n_source_individs,
                                                   n_source_individs + 1)]
        location_offset = [0] * (tc.individuals.num_rows -
                                 n_source_individs) + location_offset
        assert len(location_offset) == (tc.individuals.num_rows + 1), ("The "
                                "location_offset vector must be one longer "
                                "than the number of rows in the individuals "
                                "table on which it will be set.")

    # parents column is unused by geonomics
    # (wasn't implemented until 21/01/21, after gnx was developed;
    # tskit-dev commit 8ebfd8b2900217dab2b82d6f530c830238037fc8).
    # useful but not necessary to implement;
    # this column is also empty in msprime

    # gnx uses the metadata column to store each individual's
    # gnx Individual idx, for later matching to rows in the individuals table
    # when the table is updated after tskit's simplify algo filters individuals;
    # determine the next n indx values using the Species.max_ind_idx attribute
    gnx_individ_idxs = [*range(recip_spp.max_ind_idx+1,
                       recip_spp.max_ind_idx + n_source_individs + 1)]
    metadata=[idx.to_bytes(length=4,
                           byteorder='little') for idx in gnx_individ_idxs]
    # for msprime, make location_offset vals uniform across new Individs' rows
    if source_software == 'msprime':
        metadata_offset = [int(n) for n in np.linspace(0,
                                                   4*tc.individuals.num_rows,
                                                   tc.individuals.num_rows+1)]
    elif source_software == 'gnx':
        # same as for location_offset above, need to tack leading zeros onto
        # the metadata vector if setting columns in a gnx-generated individuals
        # table
        metadata_offset = [int(n) for n in np.linspace(0,
                                                   4*n_source_individs,
                                                   n_source_individs+1)]
        metadata_offset = [0] * (tc.individuals.num_rows -
                                 n_source_individs) + metadata_offset
        assert len(metadata_offset) == (tc.individuals.num_rows + 1), ("The "
                                "metadata_offset vector must be one longer "
                                "than the number of rows in the individuals "
                                "table on which it will be set.")

    # NOTE the set_columns() method fails because the metadata column, although
    #      supposed to be a binary data type according to the docs, is actually
    #      coded as np.int8 and so fails to convert the 4-byte binary data we
    #      are trying to provide; never ran into this problem in tha past
    #      because the .add_row() method allows the binary data for some reason
    #      and that's the only method I've used before (just adding 1
    #      individual at a time as they're born);
    #      however, somewhat conveniently, we were already planning to change
    #      every column already containing data in this table anyhow, so for
    #      now we are just:
    #           1.) double-checking the length of the original table
    #           2.) clearing the table
    #           3.) manually looping and adding new rows
    #           4.) confirming identical resulting length
    if source_software == 'msprime':
        assert (len(flags) ==
            coords.shape[0] ==
            (len(location_offset)-1) ==
            len(metadata) ==
            (len(metadata_offset)-1)), ("The 'flags', 'location', and "
                                        "'metadata' columns for the "
                                        "individuals table must all "
                                        "be the same length as the coords "
                                        "array's 0th dimension and the "
                                        "'location_offset' and "
                                        "'metadata_offset' columns must be "
                                        "one longer.")
    elif source_software == 'gnx':
        assert (len(flags) ==
                (len(location_offset)-1) ==
                (len(metadata_offset)-1)), ("The 'flags' column must be "
                            "one shorter than the 'location_offset' and "
                            "'metadata_offset' columns.")
        assert len(metadata) == coords.shape[0] == n_source_individs, ("The "
                            "'metadata' column must have one value "
                            "per coordinate pair provided.")
    original_inds_tab_len = tc.individuals.num_rows
    if source_software == 'msprime':
        assert original_inds_tab_len == coords.shape[0], ("Length "
                                             "of individuals table "
                                             "generated by msprime "
                                             "differs from length of coords "
                                             "attempting to be added to it.")
    tc.individuals.clear()
    for i, flag in enumerate(flags):
        # for gnx, add original flags and otherwise empty rows for all but the
        # final rows corresponding to the Individuals being unioned into a
        # recipient Species
        if source_software == 'gnx':
            if i < (len(flags) - n_source_individs):
                row_coords = None
                row_metadata = None
            else:
                i_adjusted = i - (len(flags) - n_source_individs)
                row_coords = coords[i_adjusted, :]
                row_metadata = metadata[i_adjusted]
        elif source_software == 'msprime':
            row_coords = coords[i, :]
            row_metadata = metadata[i]
        tc.individuals.add_row(flags=flag,
                               location=row_coords,
                               parents=None,
                               metadata=row_metadata,
                              )
    assert tc.individuals.num_rows == original_inds_tab_len, ("Reconstructed "
                                            "individuals table is not the same "
                                            "length as the original table "
                                            "generated by msprime.")


    # reformat edges table content:
    #-----------------------------

    # geonomics models all edges as being delineated by positions belonging
    # to the set [0, 0.5, 1.5, ..., L-2.5, L-1.5, L];
    # this is more complicated than the saner msprime
    # convention of using the set [0, 1, 2, ..., L-2, L-1, L],
    # but this was a decision I made based on modeling recombination
    # breakpoints halfway between loci and it is equally functional,
    # so for msprime-derived tables we need to adjust the positions
    if source_software == 'msprime':
        L = recip_spp.gen_arch.L
        left = tc.edges.left
        right = tc.edges.right
        left = np.clip(left - 0.5, a_min=0, a_max=L)
        right = np.clip([r-0.5 if r != L else r for r in right], a_min=0, a_max=L)
        tc.edges.left = left
        tc.edges.right = right

    # parent and child columns are both fine as is
    # (have double-checked that node ids are all automatically and correctly
    # updated by the tskit.TableCollection.union() method)

    # metadata column is currently unused by both msprime and gnx


    # reformat sites table content:
    #-----------------------------

    # double-check that position and ancestral state columns are identical in
    # the recipient Species' TableCollection;
    # if so, then just duplicate that table here
    # (because the only difference between the two in the metadata column,
    # which in gnx stores whether a site started as a neutral ('n')
    # or non-neutral ('t', for trait-influencing) site
    # NOTE: not running these checks for msprime-derived tables because
    #       they can lack some sites, depending on recomb rate, Ne, etc
    if source_software == 'gnx':
        for col in ['position', 'ancestral_state']:
            assert np.all(getattr(tc.sites, col) ==
                          getattr(recip_spp._tc.sites, col)), (
                                f"The {col} column in the sites table "
                                f"generated by msprime disagrees with the {col} "
                                 "column in the recipient Species' sites table.")
    tc.sites.replace_with(recip_spp._tc.sites)


    # reformat mutations table content:
    #---------------------------------

    # site, node, and parent columns are the core pieces needed to recover the
    # genotypes produced by the mutations, and we want whatever genotypes
    # msprime has simulated, so these columns should all remain untouched
    # site vals are all in interval [0, L-1]
    assert (np.all(0 <= tc.mutations.site) and
            np.all(tc.mutations.site <= (recip_spp.gen_arch.L-1))), ("The "
                        "site column in the mutations table contains values "
                        "not in the interval [0, recip_spp.gen_arch.L-1].")

    # Times for mutations created at the start of a gnx simulation, to
    # match the starting allele frequency array ('p') provided to a gnx
    # model, are artificial as they are all forcibly mutated at at once.
    # Thus, those times are NaN. Unfortunately, tskit does not all
    # mutation times to be mixed known and unknwon, so that means we must
    # set all mutation times to NaN
    time = [tskit.UNKNOWN_TIME] * len(tc.mutations.time)
    tc.mutations.time = time

    # double-check that the derived_state_offset column is just serial
    # integers from 0 to the table length, inclusive
    assert np.all(tc.mutations.derived_state_offset ==
                  np.linspace(0,
                              tc.mutations.num_rows,
                              tc.mutations.num_rows + 1)), ("The "
                "derived_state_offset column of the mutations table should "
                "contain serial integers from 0 to the table length, "
                "inclusive, but does not.")

    # metadata, metadata_offset, and metadata_schema are all currently unused
    # in both msprime and gnx


    # reformat migrations table content:
    #-----------------------------------

    # this table is not used in gnx, but if it was used in a complex msprime
    # Demography that simulated individuals for introduction into gnx then it
    # will not be empty in this TableCollection; in that case, its contents
    # will just be unioned into the otherwise empty migrations table of the 
    # recipient Species and then left as is


    # reformat populations table content:
    #-----------------------------------

    # This table is unused by gnx, but the msprime table has
    # basic population names (e.g., 'pop_0') and descriptions ('' by default,
    # but a user could specify).
    # We won't touch or do anything with this data, and the
    # TableCollection.union() method will pull it along without editing,
    # but the rows in this table will be appended to existing rows in the gnx
    # model and the nodes table will be constructed so as to match those indices

    # reformat provenances table content:
    #-----------------------------------

    # no need to do anything here; all provenances info will be retained and
    # combined when the union() method is called on the recipient Species'
    # TableCollection

    # return the gnx idxs that were set in the individuals table, for
    # downstream use
    return gnx_individ_idxs


def _get_lineage_dicts(spp, nodes, loci, t_curr, drop_before_sim=True,
                       time_before_present=True, max_time_ago=None,
                       min_time_ago=None):
    '''
    For a sample of nodes and a sample of loci, get the
    loci-sequentially ordered dict of lineage dicts
    (containing each node in a child node's lineage as the
    key and its birth-time and birth-location as a length-2 list of values

    Thus, the returned value is of the structure:
        {locus_id_1: {node_id_1: {curr_node_id: (birth_time,
                                             array([birth_loc_par_1,
                                                    birth_loc_par_2])),
                              prev_node_id: (birth_time,
                                             array([birth_loc_par_1,
                                                    birth_loc_par_2])),
                              .
                              .
                              .
                              oldest_node_id: (birth_time,
                                             array([birth_loc_par_1,
                                                    birth_loc_par_2]))},
                    .
                    .
                    .
                    node_id_last: {curr_node_id: ...
                                    ...
                                    }},
        .
        .
        .
        locus_id_last: {node_id_1: {curr_node_id: ...
                                    ...
                                    }}
        }
    '''

    # make sure the TableCollection is sorted, then get the TreeSequence
    try:
        ts = spp._tc.tree_sequence()
    except Exception:
        raise Exception(("The species' TableCollection must be sorted and "
                         "simplified before this method is called."))
    # get the tree numbers corresponding to each locus
    treenums = _get_treenums(ts, loci)
    # get the trees
    trees = ts.aslist()
    # create the output data structure
    lineage_dicts = {}
    # get the lineage_dict for each locus
    for loc, treenum in zip(loci, treenums):
        # store the lineage dicts for the current locus
        lineage_dicts_curr_loc = _get_lineage_dicts_one_tree(spp._tc,
                                    trees[treenum], nodes, t_curr,
                                    drop_before_sim=drop_before_sim,
                                    time_before_present=time_before_present,
                                    max_time_ago=max_time_ago,
                                    min_time_ago=min_time_ago)
        lineage_dicts[loc] = lineage_dicts_curr_loc
    return lineage_dicts


def _get_lineage_dicts_one_tree(tc, tree, nodes, t_curr, drop_before_sim=True,
                                time_before_present=True, max_time_ago=None,
                                min_time_ago=None):
    '''
    For a sample of nodes, and for a given tree,
    get dicts of each node's lineage nodes with their birth locs and birth
    times
    '''

    #create an output master dict
    lineage_dicts = {}
    # get each sample node's lineage dict
    for node in nodes:
        lineage = _get_lineage(tree, [node])
        lineage_dict =dict(zip(lineage, _get_lineage_times_and_locs(
                                    tc, lineage, t_curr,
                                    drop_before_sim=drop_before_sim,
                                    time_before_present=time_before_present)))

         # filter for time, if requested
        if max_time_ago is not None or min_time_ago is not None:
            # if only one time limit was set, set the other correctly, based on
            # whether or not time is expressed as time before present
            if max_time_ago is None:
                max_time_ago = np.inf
            if min_time_ago is None:
                min_time_ago = -np.inf
            # filter by time
            lineage_dict = {node: val for node, val in lineage_dict.items() if (
                            min_time_ago <= val[0] <= max_time_ago)}

        lineage_dicts[node] = lineage_dict

    return lineage_dicts


# get birth-locations and times of nodes along a pedigree
def _get_lineage_times_and_locs(tc, lineage, t_curr, drop_before_sim=True,
                                time_before_present=True):
    locs = []
    times = []
    for node in lineage:
        time = tc.nodes[node].time
        # NOTE: optionally drop all nodes from before the simulation
        # (i.e. nodes which were added to TableCollection by
        # backward-time simulation with ms); THIS IS THE DEFAULT
        # (or include all, if drop_before_sim is False)
        if (drop_before_sim and time < 0) or not drop_before_sim:
            times.append(time)
            individ = tc.nodes[node].individual
            loc = tc.individuals[individ].location
            locs.append(loc)
    # express times as time before present, if requested
    if time_before_present:
        # NOTE: need to make t_curr negative, to reflect the fact that time is
        # stored as positive integeres in Geonomics data structures but as
        # negative integers in tskit TableCollection (because they require time
        # expressed as time before present, so using positive integers would
        # not allow models to be run indefinitely, for exploratory/fun purposes
        times = [t - -t_curr for t in times]
    return(zip(times, locs))


def _get_lineage(tree, nodes):
    '''
    Recursive algorithm for getting all parents of a node
    '''
    parent = tree.parent(nodes[-1])
    if parent == tskit.NULL:
        return(nodes)
    else:
        nodes.append(parent)
        return(_get_lineage(tree, nodes))


# get a dict of the interval (i.e. tree) number
# keyed to each locus in a list of loci
def _get_treenums(ts, loci, as_dict=False):
    interval_right_ends = [t.get_interval()[1] for t in ts.trees()]
    treenums = [bisect.bisect_left(interval_right_ends, loc) for loc in loci]
    if as_dict:
        treenums = {loc: treenum for loc, treenum in zip(loci, treenums)}
    return(treenums)


# calculate a stat for the lineage corresponding to the provided lineage_dict
def _calc_lineage_stat(lin_dict, stat):
    assert stat in ['dir', 'dist', 'time', 'speed'], ("The only valid "
                                                       "statistics are: "
                                                       "direction ('dir'), "
                                                       "distance ('dist'), "
                                                       "time ('time'), "
                                                       "and speed ('speed').")
    fn_dict = {'dir': _calc_lineage_direction,
               'dist': _calc_lineage_distance,
               'time': _calc_lineage_time,
               'speed': _calc_lineage_speed
              }
    val = fn_dict[stat](lin_dict)
    return val


# calculate the angular direction of gene flow along a locus' lineage
def _calc_lineage_direction(lin_dict):
    """
    returns 'compass' directions of gene flow, with directions expressed
    as angles (in degrees) clockwise from compass north
    """
    # if the lineage has 1 or less node, don't calculate anything
    if len(lin_dict.values()) < 2:
        return
    else:
        # get x and y distances between beginning and ending points
        # (begins at the bottom of the lineage dict, furthest back in time)
        beg_loc = [*lin_dict.values()][-1][1]
        # (ends at the top of the lineage dict, in the current time step)
        end_loc = [*lin_dict.values()][0][1]
        # get the difference in the x and y axes between the lineage's
        # beginning location and its ending (i.e. current) location
        # NOTE: using range(2) will grab only the first location values,
        # which is perfect, because models with traits also save phenotypic
        # and fitness values in the individuals table's 'location' column
        x_diff, y_diff = [end_loc[i] - beg_loc[i] for i in range(2)]
        # get the counterclockwise angle, expressed in degrees
        # from the vector (X,Y) = (1,0),
        # with 0 to 180 in quadrants 1 & 2, 0 to -180 in quadrants 3 & 4
        ang = np.rad2deg(np.arctan2(y_diff, x_diff))
        # convert to all positive values, 0 - 360
        if ang < 0:
            ang += 360
        #convert to all positive clockwise angles,
        # expressed as 0 at compass north
        # (i.e. angles clockwise from vector (X,Y) = (0,1)
        ang = (-ang + 90) % 360
        return ang


# calculate the geographic distance (in cell widths)
# of gene flow along a locus' lineage
def _calc_lineage_distance(lin_dict):
    if len(lin_dict.values()) < 2:
        return
    else:
        # get x and y distances between beginning and ending points
        beg_loc = [*lin_dict.values()][0][1]
        end_loc = [*lin_dict.values()][-1][1]
        x_diff, y_diff = [end_loc[i] - beg_loc[i] for i in range(2)]
        #Pythagoras
        dist = np.sqrt(x_diff**2 + y_diff**2)
        return dist


# calculate the total time to the simulation's MRCA of a locus' lineage
def _calc_lineage_time(lin_dict):
    if len(lin_dict.values()) < 2:
        return
    else:
        beg_time = [*lin_dict.values()][0][0]
        end_time = [*lin_dict.values()][-1][0]
        time = end_time - beg_time
        return time


# calculate the speed of gene flow in a lineage (in cell widths/time steps)
def _calc_lineage_speed(lin_dict):
    if len(lin_dict.values()) < 2:
        return
    else:
        dist = _calc_lineage_distance(lin_dict)
        time = _calc_lineage_time(lin_dict)
        speed = dist/time
        return speed




#######################################################
#######################################################
#######################################################
#######################################################
#######################################################
#######################################################
#######################################################


import numpy as np
import geonomics as gnx
import msprime
import tskit
import warnings
from copy import deepcopy
from collections import Counter as C




#!/usr/bin/python
# params.py

'''
Classes and functions for creating and reading parameters files
'''

#geonomics imports
from geonomics.utils._str_repr_ import _get_str_spacing

#other imports
import os, time, datetime
import numpy as np
import pandas as pd
import copy
import re


######################################
# -----------------------------------#
# VARIABLES -------------------------#
# -----------------------------------#
######################################

    ##################
    # params strings #
    ##################

#main text-block
#STRING SLOTS:
    #%s = layers_params,
    #%s = spps_params,
    #%s = its_params,
    #%s = data_params,
    #%s = stats_params,
PARAMS = '''# %s

# This is a parameters file generated by Geonomics
# (by the gnx.make_parameters_file() function).


                   #   ::::::          :::    :: :::::::::::#
             #::::::    ::::   :::      ::    :: :: ::::::::::: ::#
          #:::::::::     ::            ::   :::::::::::::::::::::::::#
        #::::::::::                      :::::::::: :::::: ::::::::  ::#
      #  : ::::  ::                    ::::  : ::    :::::::: : ::  :    #
     # GGGGG :EEEE: OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS #
    # GG     EE    OO   OO  NNN  NN  OO   OO  MM   MM   II   CC     SS     #
    # GG     EE   OO     OO NN N NN OO     OO MMM MMM   II   CC     SSSSSS #
    # GG GGG EEEE OO     OO NN  NNN OO     OO MM M MM   II   CC         SS #
    # GG   G EE    OO   OO  NN   NN  OO   OO  MM   MM   II   CC        SSS #
     # GGGGG :EEEE: OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS #
      #    : ::::::::               :::::::::: ::              ::  :   : #
        #:    :::::                    :::::: :::             :::::::  #
          #    :::                      :::::  ::              ::::: #
             #  ::                      ::::                      #
                   #                                        #
                      #  :: ::    :::             #


params = {
#-----------------------------------------------------------------------------#

#-----------------#
#--- LANDSCAPE ---#
#-----------------#
    'landscape': {

    #------------#
    #--- main ---#
    #------------#
        'main': {
            #x,y (a.k.a. j,i) dimensions of the Landscape
            'dim':                      (20,20),
            #x,y resolution of the Landscape
            'res':                      (1,1),
            #x,y coords of upper-left corner of the Landscape
            'ulc':                      (0,0),
            #projection of the Landscape
            'prj':                      None,
            }, # <END> 'main'

    #--------------#
    #--- layers ---#
    #--------------#
        'layers': {
%s


    #### NOTE: Individual Layers' sections can be copy-and-pasted (and
    #### assigned distinct keys and names), to create additional Layers.


            } # <END> 'layers'

        }, # <END> 'landscape'


#-----------------------------------------------------------------------------#

#-----------------#
#--- COMMUNITY ---#
#-----------------#
    'comm': {

        'species': {
%s


    #### NOTE: individual Species' sections can be copy-and-pasted (and
    #### assigned distinct keys and names), to create additional Species.


            }, # <END> 'species'

        }, # <END> 'comm'


#-----------------------------------------------------------------------------#

#-------------#
#--- MODEL ---#
#-------------#
    'model': {
        #total Model runtime (in timesteps)
        'T':            100,
        #min burn-in runtime (in timesteps)
        'burn_T':       30,
        #seed number
        'num':          None,

%s
%s
%s
        } # <END> 'model'

    } # <END> params
'''

#block for lyr params
#STRING SLOTS: 
    #%s = lyr_name,
    #%i = lyr_num,
    #%s = lyr_type_params,
    #%s = lyr_change_params,
    #%i = lyr_num,
LYR_PARAMS = '''
            #layer name (LAYER NAMES MUST BE UNIQUE!)
            %s: {

        #-------------------------------------#
        #--- layer num. %i: init parameters ---#
        #-------------------------------------#

                #initiating parameters for this layer
                'init': {
%s
                    }, # <END> 'init'
%s
                }, # <END> layer num. %i
'''

#the block of random-layer parameters
RAND_LYR_PARAMS = '''
                    #parameters for a 'random'-type Layer
                    'random': {
                        #number of random points
                        'n_pts':                        500,
                        #interpolation method {'linear', 'cubic', 'nearest'}
                        'interp_method':                'linear',

                        }, # <END> 'random'
'''

#the block of defined-layer parameters
DEFINED_LYR_PARAMS = '''
                    #parameters for a 'defined'-type Layer
                    'defined': {
                        #raster to use for the Layer
                        'rast':                   np.ones((20,20)),
                        #point coordinates
                        'pts':                    None,
                        #point values
                        'vals':                   None,
                        #interpolation method {None, 'linear', 'cubic',
                        #'nearest'}
                        'interp_method':          None,

                        }, # <END> 'defined'
'''

#the block of file-layer parameters
FILE_LYR_PARAMS = '''
                    #parameters for a 'file'-type Layer
                    'file': {
                        #</path/to/file>.<ext>
                        'filepath':                     '/PATH/TO/FILE.EXT',
                        #minimum value to use to rescale the Layer to [0,1]
                        'scale_min_val':                None,
                        #maximum value to use to rescale the Layer to [0,1]
                        'scale_max_val':                None,
                        #decimal precision to use for coord-units (ulc & res)
                        'coord_prec':                   5,
                        #units of this file's variable
                        'units':                        None,

                        }, # <END> 'file'
'''

#the block of nlmpy-layer parameters
NLMPY_LYR_PARAMS = '''
                    #parameters for an 'nlmpy'-type Layer
                    'nlmpy': {
                        #nlmpy function to use the create this Layer
                        'function':                 'mpd',
                        #number of rows (MUST EQUAL LANDSCAPE DIMENSION y!)
                        'nRow':                     20,
                        #number of cols (MUST EQUAL LANDSCAPE DIMENSION x!)
                        'nCol':                     20,
                        #level of spatial autocorrelation in element values
                        'h':                        1,

                        }, # <END> 'nlmpy'
'''

#block of layer-change parameters
#STRING SLOTS:
    #%i = lyr_num,
    #%s = change events,
LYR_CHANGE_PARAMS = '''
            #---------------------------------------#
            #--- layer num. %i: change parameters ---#
            #---------------------------------------#

                #landscape-change events for this Layer
                'change': {
%s

                    }, # <END> 'change'
'''

#block for params in a single layer-change event
#STRING SLOTS:
    #%i = lyr_change_event_num,
    #%i = lyr_change_event_num,
LYR_CHANGE_EVENT_PARAMS = '''
                    %i: {
                        #array or file for final raster of event, or directory
                        #of files for each stepwise change in event
                        'change_rast':      '/PATH/TO/FILE.EXT',
                        #starting timestep of event
                        'start_t':          49,
                        #ending timestep of event
                        'end_t':            99,
                        #number of stepwise changes in event
                        'n_steps':          5,
                        }, # <END> event %i'''

#block of species params
#STRING SLOTS:
    #%s = spp_name,
    #%i = spp_num,
    #%s = msprime_init_params,
    #%i = spp_num,
    #%i = spp_num,
    #%i = spp_num,
    #%s = move_surf_params,
    #%s = disp_surf_params,
    #%s = genome_params,
    #%s = change_params,
    #%i = spp_num,
SPP_PARAMS = '''
            #species name (SPECIES NAMES MUST BE UNIQUE!)
            %s: {

            #-----------------------------------#
            #--- spp num. %i: init parameters ---#
            #-----------------------------------#

                'init': {
                    #starting number of individs
                    'N':                250,
                    #carrying-capacity Layer name
                    'K_layer':          'lyr_0',
                    #multiplicative factor for carrying-capacity layer
                    'K_factor':         1,%s
                    }, # <END> 'init'

            #-------------------------------------#
            #--- spp num. %i: mating parameters ---#
            #-------------------------------------#

                'mating'    : {
                    #age(s) at sexual maturity (if tuple, female first)
                    'repro_age':                0,
                    #whether to assign sexes
                    'sex':                      False,
                    #ratio of males to females
                    'sex_ratio':                1/1,
                    #intrinsic growth rate
                    'R':                        0.5,
                    #intrinsic birth rate (MUST BE 0<=b<=1)
                    'b':                        0.2,
                    #expectation of distr of n offspring per mating pair
                    'n_births_distr_lambda':    1,
                    #whether n births should be fixed at n_births_dist_lambda
                    'n_births_fixed':           True,
                    #radius of mate-search area (None, for panmixia)
                    'mating_radius':            10,
                    #whether individs should choose nearest neighs as mates
                    'choose_nearest_mate':        False,
                    #whether mate-choice should be inverse distance-weighted
                    'inverse_dist_mating':      False,
                    }, # <END> 'mating'

            #----------------------------------------#
            #--- spp num. %i: mortality parameters ---#
            #----------------------------------------#

                'mortality'     : {
                    #maximum age
                    'max_age':                      None,
                    #min P(death) (MUST BE 0<=d_min<=1)
                    'd_min':                        0,
                    #max P(death) (MUST BE 0<=d_max<=1)
                    'd_max':                        1,
                    #width of window used to estimate local pop density
                    'density_grid_window_width':    None,
                    }, # <END> 'mortality'

            #---------------------------------------#
            #--- spp num. %i: movement parameters ---#
            #---------------------------------------#

                'movement': {
                    #whether or not the species is mobile
                    'move':                                 True,
                    #mode of distr of movement direction
                    'direction_distr_mu':                   0,
                    #concentration of distr of movement direction
                    'direction_distr_kappa':                0,
                    #1st param of distr of movement distance
                    'movement_distance_distr_param1':       0.01,
                    #2nd param of distr of movement distance
                    'movement_distance_distr_param2':       0.5,
                    #movement distance distr to use ('lognormal','levy','wald')
                    'movement_distance_distr':              'lognormal',
                    #1st param of distr of dispersal distance
                    'dispersal_distance_distr_param1':      -1,
                    #2nd param of distr of dispersal distance
                    'dispersal_distance_distr_param2':      0.05,
                    #dispersal distance distr to use ('lognormal','levy','wald')
                    'dispersal_distance_distr':             'lognormal',%s%s
                    },    # <END> 'movement'

%s
%s
                }, # <END> spp num. %i
'''

# block for params to initiate model with
# individuals drawn from msprime simulations
MSPRIME_INIT_PARAMS = '''
                    # params for 1+ msprime pops for sampling starting individs
                    'msprime':        {%s
                        },
'''

MSPRIME_POP_PARAMS = '''
                        # index of msprime source pop
                        %i:  {
                            # number of individs to sample from pop
                            100: {
                                 # 1x2 coord pair or nx2 coord pairs of individs
                                 'coords':              [0, 0],
                                 # valid kwargs dict for Model.add_individuals
                                 # 'source_msprime_params' argument...
                                 'recomb_rate':         0.5,
                                 'mut_rate':            0.0001,
                                 'demography':          None,
                                 'population_size':     None,
                                 'ancestry_model':      None,
                                 'random_seed':         None,
                                 }
                            },

'''

#block for movement-surface params
MOVE_SURF_PARAMS = '''
                    'move_surf'     : {
                        #move-surf Layer name
                        'layer':                'lyr_0',
                        #whether to use mixture distrs
                        'mixture':              True,
                        #concentration of distrs
                        'vm_distr_kappa':       12,
                        #length of approximation vectors for distrs
                        'approx_len':           5000,
                        }, # <END> 'move_surf'
'''

#block for dispersal-surface params
DISP_SURF_PARAMS = '''
                    'disp_surf'     : {
                        #disp-surf Layer name
                        'layer':                'lyr_0',
                        #whether to use mixture distrs
                        'mixture':              True,
                        #concentration of distrs
                        'vm_distr_kappa':       12,
                        #length of approximation vectors for distrs
                        'approx_len':           5000,
                        }, # <END> 'disp_surf'
'''

#block for genome params
#STRING SLOTS:
    #%i = spp_num,
    #
    #%s = traits_params,
GENOME_PARAMS = '''
            #---------------------------------------------------#
            #--- spp num. %i: genomic architecture parameters ---#
            #---------------------------------------------------#

                'gen_arch': {
                    #file defining custom genomic arch
                    'gen_arch_file':            %s,
                    #num of loci
                    'L':                        100,
                    #fixed starting allele freq; None/False -> rand; True -> 0.5
                    'start_p_fixed':            0.5,
                    #whether to start neutral locus freqs at 0
                    'start_neut_zero':          False,
                    #genome-wide per-base neutral mut rate (0 to disable)
                    'mu_neut':                  0,
                    #genome-wide per-base deleterious mut rate (0 to disable)
                    'mu_delet':                 0,
                    #shape of distr of deleterious effect sizes
                    'delet_alpha_distr_shape':  0.2,
                    #scale of distr of deleterious effect sizes
                    'delet_alpha_distr_scale':  0.2,
                    #alpha of distr of recomb rates
                    'r_distr_alpha':            0.5,
                    #beta of distr of recomb rates
                    'r_distr_beta':             None,
                    #whether loci should be dominant (for allele '1')
                    'dom':                      False,
                    #whether to allow pleiotropy
                    'pleiotropy':               False,
                    #custom fn for drawing recomb rates
                    'recomb_rate_custom_fn':    None,
                    #number of recomb paths to hold in memory
                    'n_recomb_paths_mem':       int(1e4),
                    #total number of recomb paths to simulate
                    'n_recomb_paths_tot':       int(1e5),
                    #num of crossing-over events (i.e. recombs) to simulate
                    'n_recomb_sims':            10_000,
                    #whether to generate recombination paths at each timestep
                    'allow_ad_hoc_recomb':      False,
                    #whether to jitter recomb bps, to correctly track num_trees
                    'jitter_breakpoints':       False,
                    #whether to save mutation logs
                    'mut_log':                  False,
                    #whether to use tskit (to record full spatial pedigree)
                    'use_tskit':                True,
                    #time step interval for simplication of tskit tables
                    'tskit_simp_interval':      100,

%s
                    }, # <END> 'gen_arch'
'''

#block for traits params
#STRING SLOTS:
    #%s = multi_trait_params,
TRTS_PARAMS = '''
                    'traits': {
%s

    #### NOTE: Individual Traits' sections can be copy-and-pasted (and
    #### assigned distinct keys and names), to create additional Traits.


                        }, # <END> 'traits'
'''

#block for trait params
#STRING SLOTS:
    #%i = trait_num,
    #%s = trait_num,
    #%i = trait_num,
TRT_PARAMS = '''
                        #-------------------------#
                        #---trait %i parameters ---#
                        #-------------------------#
                        #trait name (TRAIT NAMES MUST BE UNIQUE!)
                        %s: {
                            #trait-selection Layer name
                            'layer':                'lyr_0',
                            #phenotypic selection coefficient
                            'phi':                  0.05,
                            #number of loci underlying trait
                            'n_loci':               1,
                            #mutation rate at loci underlying trait
                            'mu':                   0,
                            #mean of distr of effect sizes
                            'alpha_distr_mu' :      0.1,
                            #variance of distr of effect size
                            'alpha_distr_sigma':    0,
                            #max allowed magnitude for an alpha value
                            'max_alpha_mag':        None,
                            #curvature of fitness function
                            'gamma':                1,
                            #whether the trait is universally advantageous
                            'univ_adv':             False
                            }, # <END> trait %i
'''

#block of spp_change params
#STRING SLOTS:
    #%i = spp_num,
    #%s = dem_and-or_param_change_params_str,
SPP_CHANGE_PARAMS = '''
            #-------------------------------------#
            #--- spp num. %i: change parameters ---#
            #-------------------------------------#

                'change': {
%s
                        } # <END> 'change'
'''

#block for a series of demographic-change events
#STRING SLOTS:
    #%s = multi_dem_change_event_params,
SPP_DEM_CHANGE_EVENTS_PARAMS = '''
                    #-------------------------------------#
                    #--- demographic change parameters ---#
                    #-------------------------------------#
                    'dem': {
%s


    #### NOTE: Individual demographic change events' sections can be
    #### copy-and-pasted (and assigned distinct keys and names), to create
    #### additional events.


                        }, # <END> 'dem'
'''

#block for parameters for a single demographic-change event
#STRING SLOTS:
    #%i = dem_change_event_num,
    #%i = dem_change_event_num,
SPP_DEM_CHANGE_EVENT_PARAMS = '''
                        %i: {
                            #kind of event {'monotonic', 'stochastic',
                            #'cyclical', 'custom'}
                            'kind':             'monotonic',
                            #starting timestep
                            'start_t':          49,
                            #ending timestep
                            'end_t':            99,
                            #rate, for monotonic change
                            'rate':             1.02,
                            #interval of changes, for stochastic change
                            'interval':         1,
                            #distr, for stochastic change {'uniform', 'normal'}
                            'distr':            'uniform',
                            #num cycles, for cyclical change
                            'n_cycles':         10,
                            #min & max sizes, for stochastic & cyclical change
                            'size_range':       (0.5, 1.5),
                            #list of timesteps, for custom change
                            'timesteps':        [50, 90, 95],
                            #list of sizes, for custom change
                            'sizes':            [2, 5, 0.5],
                            }, # <END> event %i

'''

#block for a series of life-history parameter-change events
SPP_PARAM_CHANGE_PARAMS = '''
                    #--------------------------------------#
                    #--- life-history change parameters ---#
                    #--------------------------------------#
                    'life_hist': {
                        #life-history parameter to change
                        '<life_hist_param>': {
                            #list of timesteps
                            'timesteps':        [],
                            #list of values
                            'vals':             [],
                                }


    #### NOTE: Individual life-history paramter change events' sections can be
    #### copy-and-pasted (and assigned distinct keys and names), to create
    #### additional events.


                            }, # <END> 'life_hist'
'''

#block for model iterations params
ITS_PARAMS = '''
        #-----------------------------#
        #--- iterations parameters ---#
        #-----------------------------#
        'its': {
            #num iterations
            'n_its':            1,
            #whether to randomize Landscape each iteration
            'rand_landscape':   False,
            #whether to randomize Community each iteration
            'rand_comm':        False,
            #whether to randomize GenomicArchitectures each iteration
            'rand_genarch':     True,
            #whether to burn in each iteration
            'repeat_burn':      False,
            }, # <END> 'iterations'
'''

#block for model data-collection params
DATA_PARAMS = '''
        #----------------------------------#
        #--- data-collection parameters ---#
        #----------------------------------#
        'data': {
            'sampling': {
                #sampling scheme {'all', 'random', 'point', 'transect'}
                'scheme':               'random',
                #sample size at each point, for point & transect sampling
                'n':                    250,
                #coords of collection points, for point sampling
                'points':               None,
                #coords of transect endpoints, for transect sampling
                'transect_endpoints':   None,
                #num points along transect, for transect sampling
                'n_transect_points':    None,
                #collection radius around points, for point & transect sampling
                'radius':               None,
                #when to collect data
                'when':                 None,
                #whether to save current Layers when data is collected
                'include_landscape':    False,
                #whether to include fixed loci in VCF files
                'include_fixed_sites':  False,
                },
            'format': {
                #format for genetic data {'vcf', 'fasta'}
                'gen_format':           ['vcf', 'fasta'],
                #format for vector geodata {'csv', 'shapefile', 'geojson'}
                'geo_vect_format':      'csv',
                #format for raster geodata {'geotiff', 'txt'}
                'geo_rast_format':      'geotiff',
                #format for file listing non-neut loci (if 'csv', will be saved)
                'nonneut_loc_format': None,
                },
            }, #<END> 'data'
'''

#block for model stats-calculation params
STATS_PARAMS = '''
        #-----------------------------------#
        #--- stats-collection parameters ---#
        #-----------------------------------#
        'stats': {
            #number of individs at time t
            'Nt': {
                #whether to calculate
                'calc':     True,
                #calculation frequency (in timesteps)
                'freq':     1,
                },
            #heterozgosity
            'het': {
                #whether to calculate
                'calc':     True,
                #calculation frequency (in timesteps)
                'freq':     5,
                #whether to mean across sampled individs
                'mean': False,
                },
            #minor allele freq
            'maf': {
                #whether to calculate
                'calc':     True,
                #calculation frequency (in timesteps)
                'freq':     5,
                },
            #mean fitness
            'mean_fit': {
                #whether to calculate
                'calc':     True,
                #calculation frequency (in timesteps)
                'freq':     5,
                },
            #linkage disequilibirum
            'ld': {
                #whether to calculate
                'calc':     False,
                #calculation frequency (in timesteps)
                'freq':     100,
                },
            }, # <END> 'stats'
'''

######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

#a _DynAttrDict dict class with k:v pairs as dynamic attributes
class _DynAttrDict(dict):
    def __getattr__(self, item):
        return self[item]
    def __dir__(self):
        return super().__dir__() + [str(k) for k in self.keys()]
    def __deepcopy__(self, memo):
        return _DynAttrDict(copy.deepcopy(dict(self)))

#a ParametersDict class (which is just a recursion the _DynAttrDict over the
#whole params dict, to make all its levels dicts with dynamic attributes, 
#i.e indexable by dot notation and responsive to tab completion)
class ParametersDict(_DynAttrDict):
    def __init__(self, params):
        params_dict = _make_params_dict(params)
        self.update(params)

    #re-enable deepcopy, because the class inherits from a dict
    def __deepcopy__(self, memo):
        return ParametersDict(copy.deepcopy(dict(self)))

    #define the __str__ and __repr__ special methods
    def __str__(self):
        #get a string representation of the class
        type_str = str(type(self))
        #get the model name str
        name_str = "Model name:%s%s"
        name_str = name_str % (_get_str_spacing(name_str), self.model.name)
        #concatenate the strings
        tot_str = '\n'.join([type_str, name_str])
        return tot_str

    def __repr__(self):
       repr_str = self.__str__()
       return repr_str


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

#function to create the lyrs-params section of a params file
def _make_lyrs_params_str(lyrs=1):
    #create an empty list, to be filled with one params string per lyr
    lyrs_params_list = []
    #if lyrs is an integer, create a string of identical parameter sections
    if type(lyrs) is int:
        #assert that it's an integer greater than 0
        assert lyrs > 0, ("The number of Layers to be created must be a "
        "positive integer.")
        #for each lyr
        for i in range(lyrs):
            #use lyr-type 'random'
            type_params = RAND_LYR_PARAMS
            #add no change params (i.e. a zero-length string)
            change_params = ''
            #create the lyr_params str
            lyr_params_str = LYR_PARAMS % ("'lyr_%i'" % i, i,
                                               type_params, change_params, i)
            #append it to the list
            lyrs_params_list.append(lyr_params_str)

    #or if lyrs is a list of dicts, then create individually customized
    #params sections for each Layer
    elif type(lyrs) is list:
        #assert that each item in the list is a dict
        assert False not in [type(item) is dict for item in lyrs], ("All "
            "items in the argument 'layers' must be of type dict if it is "
            "provided as a list.")
        assert False not in [type(item) is dict for item in lyrs], ("If the "
            "'layers' argument is a list then it must contain only "
            "objects of type dict.")
        #create lookup dicts for the params strings for different lyr params
        #sections
        lyr_type_params_str_dict = {'random': RAND_LYR_PARAMS,
                                      'defined': DEFINED_LYR_PARAMS,
                                      'file': FILE_LYR_PARAMS,
                                      'nlmpy': NLMPY_LYR_PARAMS,
                                     }
        #for each lyr
        for i, lyr_dict in enumerate(lyrs):
            #assert that the 'type' value is valid
            if 'type' in [*lyr_dict]:
                assert lyr_dict['type'] in ['random', 'defined', 'file',
                    'nlmpy'], ("The value provided for the 'type' of Layer "
                    "%i is invalid. Valid values include: ['random', "
                    "'defined', 'file', 'nlmpy'].") % i
                #get the type params for this lyr
                lyr_type = lyr_dict['type']
            else:
                lyr_type = 'random'
            type_params = lyr_type_params_str_dict[lyr_type]
            #assert that the 'change' value is valid
            if 'change' in [*lyr_dict]:
                assert (isinstance(lyr_dict['change'], bool)
                        or (isinstance(lyr_dict['change'], int)
                           and lyr_dict['change'] > 0)), ("The value "
                    "provided for the 'change' value of Layer %i is invalid. "
                    "Value must be either a boolean (to create a single "
                    "change event), or a positive int (to create a series "
                    "of that many change events).") % i
                #get the change params for this lyr
                lyr_change = lyr_dict['change']
            else:
                lyr_change = False
            #get the LYR_CHANGE_PARAMS, if change argument is True or int > 0
            if lyr_change is not False:
                change_params = LYR_CHANGE_PARAMS
                #get the necessary number of change events' params strings
                events_params = '\n'.join([LYR_CHANGE_EVENT_PARAMS % (n,
                    n) for n in range(lyr_change)])
            else:
                change_params = ''
            if change_params != '':
                change_params = change_params % (i, events_params)
            #create the lyr_params str for this Layer
            lyr_params_str = LYR_PARAMS % ("'lyr_%i'" % i, i,
                                               type_params, change_params, i)
            #append it to the list
            lyrs_params_list.append(lyr_params_str)

    #join the whole list into one str
    lyrs_params_str = '\n'.join(lyrs_params_list)
    return lyrs_params_str


#function to create the spps-params section of a params file
def _make_species_params_str(species=1):
    assert (isinstance(species, int) or
            (isinstance(species, list) and
             np.all([isinstance(item, dict) for item in species]))), (
                 "The 'species' param must be either an int (the number "
                 "of species parameters sections to include in the file) "
                 "or a list of dicts (one dict per species parameters "
                 "section to be included in the file, with keys and "
                 "values stipulating the aspects of the species to be "
                 "parameterized).")
    #create an empty list, to be filled with one params string per spp
    spps_params_list = []
    #if species is an integer, create a string of identical parameter sections
    if type(species) is int:
        #assert that it's an integer greater than 0
        assert species > 0, ("The number of Species to be created "
        "must be a positive integer.")
                #for each spp
        for i in range(species):
            #use genome params, but with no traits (i.e. string-format with a
            #zero-length str)
            genome_params = GENOME_PARAMS % (i, 'None', '')
            #add no change params
            change_params = ''
            #create the spp_params str, with no move_surf or disp_surf params
            spp_params_str = SPP_PARAMS % ("'spp_%i'" % i,
                                           i,
                                           '',
                                           i,
                                           i,
                                           i,
                                           '',
                                           '',
                                           genome_params,
                                           change_params,
                                           i,
                                          )
            #append to the spps_params_list
            spps_params_list.append(spp_params_str)

    #or if species is a list of dicts, then create individually customized
    #params sections for each Species
    elif type(species) is list:
        #create an empty list, to which any species who should have custom
        #gen-arch files created will be appended
        #assert that each item in the list is a dict
        assert False not in [type(item) is dict for item in species], ("If"
            " the 'species' argument is a list then it must contain only "
            "objects of type dict.")
        #create a lookup dict for the params strings for different spp params
        #sections
        params_str_dict = {
                        'msprime': {True: MSPRIME_INIT_PARAMS},
                        'move_surf': {True: MOVE_SURF_PARAMS},
                        'disp_surf': {True: DISP_SURF_PARAMS},
                        'genome': {True: GENOME_PARAMS,
                                   'custom': GENOME_PARAMS},
                        'dem_change': {True: SPP_DEM_CHANGE_EVENT_PARAMS},
                        'dem_events': {True: SPP_DEM_CHANGE_EVENTS_PARAMS},
                        'param_change': {True: SPP_PARAM_CHANGE_PARAMS},

                            }
        [v.update({False: ''}) for v in params_str_dict.values()]
        #for each spp
        for i, spp_dict in enumerate(species):
            #assert that the argument values are valid
            bool_args= ['movement_surface', 'genomes', 'parameter_change']
            int_args = ['n_traits', 'demographic_change', 'msprime_source_pops']
            for arg in bool_args:
                if arg in [*spp_dict]:
                    assert (type(spp_dict[arg]) is bool
                        or arg == 'genomes' and spp_dict[arg] == 'custom'), (
                        "The '%s' key in "
                        "each Species' dictionary must contain a "
                        "boolean value. But dict number %i in the "
                        "'species' argument contains a non-boolean "
                        "and otherwise invalid value.") % (arg, i)
            for arg in int_args:
                if arg in [*spp_dict]:
                    assert type(spp_dict[arg]) is int, ("The '%s' "
                        "key in each Species' dictionary must contain an "
                        "integer value. But dict number %i in the "
                        "'species' argument contains a non-integer "
                        "value:\n\n\t" "'%s': %s ") % (arg, i, arg,
                                                            str(spp_dict[arg]))
                    int_arg_str_fmt_dict = {'n_traits':'Traits',
                            'demographic_change': 'demographic change events',
                            'msprime_source_pops': 'msprime source populations',
                                           }
                    assert spp_dict[arg] >= 0, ("The number of %s to "
                                    "be created must be 0 or a positive "
                                    "integer.") % (int_arg_str_fmt_dict[arg])
            # get the msprime source pops params, if required
            if 'msprime_source_pops' in [*spp_dict]:
                print('making msprime params')
                n_pops = spp_dict['msprime_source_pops']
                msprime_params = params_str_dict['msprime'][n_pops>0]
                if n_pops > 0:
                    msprime_params = msprime_params % ('\n'.join([
                        MSPRIME_POP_PARAMS % i for i in range(n_pops)]))
            else:
                msprime_params = ''
            #get the movement surf and dispersal surf params, if required
            if 'movement_surface' in [*spp_dict]:
                ms_arg = spp_dict['movement_surface']
                move_surf_params = params_str_dict['move_surf'][ms_arg]
            else:
                move_surf_params = ''
            if 'dispersal_surface' in [*spp_dict]:
                ds_arg = spp_dict['dispersal_surface']
                disp_surf_params = params_str_dict['disp_surf'][ds_arg]
            else:
                disp_surf_params = ''
            #get the genome params, if required
            if 'genomes' in [*spp_dict] and spp_dict['genomes'] in [True,
                                                                    'custom']:
                #if this species should have a custom gen_arch_file made
                if spp_dict['genomes'] == 'custom':
                    gen_arch_file_str = ("'%%%%GEN_ARCH_FILE_STR%%%%_spp-%i_"
                                                            "gen_arch.csv'")
                    gen_arch_file_str = gen_arch_file_str % i
                    #make a tmp gen_arch_file for this spp
                    tmp_gen_arch_filename = '%i_%s.tmp' % (i,
                        str(np.random.randint(0, 10000)).zfill(5))
                    _make_custom_genomic_architecture_file(
                                                    tmp_gen_arch_filename)
                else:
                    gen_arch_file_str = 'None'
                #if this species' genomes should have traits
                if 'n_traits' in [*spp_dict] and spp_dict['n_traits'] > 0:
                    #get a list of params strings of length equal
                    #to the number of traits it should have
                    trait_params_list = []
                    for trt in range(spp_dict['n_traits']):
                        trait_params = TRT_PARAMS % (trt,
                                                "'trait_%i'" % trt, trt)
                        trait_params_list.append(trait_params)
                    #get the traits_params_str
                    traits_params = TRTS_PARAMS
                    #join the list into a single str and insert it into the
                    #traits_params_str
                    traits_params = traits_params % ''.join(trait_params_list)
                else:
                    traits_params = ''
                genome_params = params_str_dict['genome'][spp_dict['genomes']]
                genome_params = genome_params % (i, gen_arch_file_str,
                                                                traits_params)
            #or get empty str
            else:
                genome_params = ''
            #get the spp-change params (if either dem or param changes are
            #required
            if (('demographic_change' in [*spp_dict]
                 and spp_dict['demographic_change'])
                or ('parameter_change' in [*spp_dict]
                    and spp_dict['parameter_change'])):
                #create an empty string to tack either/both section(s) onto
                change_series_str = ''
                #tack on the dem-change events params str, if required
                if ('demographic_change' in [*spp_dict]
                    and spp_dict['demographic_change'] > 0):
                    dem_change_event_params_list = []
                    for n in range(spp_dict['demographic_change']):
                        params_str = params_str_dict['dem_change'][True]
                        dem_change_event_params_list.append(
                                                        params_str % (n, n))
                    events_series = ''.join(dem_change_event_params_list)
                    events_params_str = params_str_dict['dem_events'][True]
                    events_params_str = events_params_str % events_series
                    change_series_str = change_series_str + events_params_str
                #tack on the param-change params, if required
                if 'parameter_change' in [*spp_dict]:
                    param_change_arg = spp_dict['parameter_change']
                    param_change_params_str = (
                    params_str_dict['param_change'][param_change_arg])
                    events_params_str=events_params_str+param_change_params_str
                change_params = SPP_CHANGE_PARAMS
                change_params = change_params % (i, events_params_str)
            #or get empty str
            else:
                change_params = ''
            #get the overall spp params str for this spp
            spp_params_str = SPP_PARAMS % ("'spp_%i'" % i,
                                           i,
                                           msprime_params,
                                           i,
                                           i,
                                           i,
                                           move_surf_params,
                                           disp_surf_params,
                                           genome_params,
                                           change_params,
                                           i,
                                          )
            #append to the spps_params_list
            spps_params_list.append(spp_params_str)
    #join the whole list into one str
    spps_params_str = ''.join(spps_params_list)
    return spps_params_str


#function to create the data- and  stats-params sections of params file
#TODO: Add option for the argument to _make_parameters_file() to list the stats 
#to be calculated??
def _make_model_params_strs(section, arg=None):
    #assert the value of arg is valid
    assert arg in [True, False, None], ("The value of the '%s' argument "
        "provided for the model is not valid values. Value must be either a "
        "boolean or None.") % section
    if arg in [False, None]:
        return ''
    else:
        if section == 'data':
            return DATA_PARAMS
        elif section == 'stats':
            return STATS_PARAMS


#function to create a default params file, to be filled in by the user
def _make_params_file(filepath=None, layers=1, species=1, data=None,
        stats=None):
    '''<see docstring in gnx.make_parameters_file>'''
    lyrs_params_str = _make_lyrs_params_str(lyrs = layers)
    spps_params_str= _make_species_params_str(species = species)
    data_params_str = _make_model_params_strs('data', arg = data)
    stats_params_str = _make_model_params_strs('stats', arg = stats)
    #TODO DECIDE IF THIS SHOULD BE MADE OPTIONAL IN SOME WAY
    its_params_str = ITS_PARAMS
    #get the filepath
    if filepath is None:
        #get a string of the date and time
        datetime_str = time.strftime("%d-%m-%Y_%H:%M:%S", time.localtime())
        #and add a default filename
        filepath = 'GNX_params_%s.py' % datetime_str
    #check the filepath is pointed somewhere valid
    assert (os.path.isdir(os.path.split(filepath)[0])
            or os.path.split(filepath)[0] == ''), ("The filepath to which to "
            "write the parameters file does not point to a valid directory.")
    #coerce the file to a .py extension if it is not already provided
    filepath = os.path.splitext(filepath)[0] + '.py'
    #create the full params-file string
    file_str = PARAMS % (os.path.split(filepath)[1], lyrs_params_str,
        spps_params_str, its_params_str, data_params_str, stats_params_str)

    #add the gen_arch_file name and then create the files, if needed
    if re.search('%%GEN_ARCH_FILE_STR%%', file_str):
        #get the file prefix
        gen_arch_file_prefix = os.path.splitext(os.path.split(filepath)[1])[0]
        #rename the tmp files that were created
        tmp_files = [f for f in os.listdir('.') if (
            os.path.splitext(f)[1] == '.tmp')]
        for tmp_file in tmp_files:
            os.rename(tmp_file, gen_arch_file_prefix +
                '_spp-%s_gen_arch.csv' % (tmp_file.split('_')[0]))
        #replace the standin pattern with the file prefix
        file_str = re.sub('%%GEN_ARCH_FILE_STR%%', gen_arch_file_prefix,
                                                                file_str)

    #write the file_str to a "GNX_params_<datetime>.py" file
    with open(filepath, 'w') as f:
        f.write(file_str)


#function to recurse over the params dictionary 
#and return it as a Parameters_Dict object (i.e. a
#dict with k:v pairs as dynamic attributes)
def _make_params_dict(params):
    for k, v in params.items():
        method_names = ['clear', 'copy', 'fromkeys', 'get', 'items', 'keys',
                        'pop', 'popitem', 'setdefault', 'update', 'values']
        assert k not in method_names, ('The key "%s" in your params '
            'file is disallowed because it would clobber a Python method. '
            'Please edit name.\n\tNOTE: It holds the following value:'
            '\n%s' % (str(k), str(v)))
        if isinstance(v, dict):
            params[k] = _make_params_dict(params[k])
    params = _DynAttrDict(params)
    return(params)


#read a params file and return a ParametersDict object
def _read_params_file(filepath):
    #create a namespace to read the params dict into
    ns = {}
    #read and execute the file (to create a plain dict called 'params')
    exec(open(filepath, 'r').read(), ns)
    #get the params object from the namespace ns
    params = ns['params']
    #turn the params dict into a ParametersDict object
    params = ParametersDict(params)
    #set the model's name
    if 'name' in params['model'] and params['model']['name'] is not None:
        name = params['model']['name']
    else:
        #get the filename (minus path and extension) as the model name
        name = os.path.splitext(os.path.split(filepath)[-1])[0]
    params.model['name'] = name
    return(params)


#create a an empty custom gen-arch file for a species 
#(will be called if 'genomes':'custom' is a k:v pair in a 
#species dict fed into _make_paramters_file's 'species' argument
def _make_custom_genomic_architecture_file(filepath):
    #create the dataframe for the CSV file
    cols = ('locus', 'p', 'dom', 'r', 'trait', 'alpha')
    row0 = ([0], [np.nan], [np.nan], [np.nan], [np.nan], [np.nan])
    df_dict = dict(zip(cols, row0))
    df = pd.DataFrame.from_dict(df_dict)
    #write it to file, without the index
    df.to_csv(filepath, index = False)


